import copy
import datetime
import functools
import getpass
import hashlib
import os
import re
import time

import appdirs
import mementos
import pandas as pd
from pyquery import PyQuery as pq

import sportsref


# TODO: move PSFConstants and GPFConstants to appdirs cache dir
def switch_to_dir(dirPath):
    """
    Decorator that switches to given directory before executing function, and
    then returning to orignal directory.
    """

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            orig_cwd = os.getcwd()
            os.chdir(dirPath)
            ret = func(*args, **kwargs)
            os.chdir(orig_cwd)
            return ret

        return wrapper

    return decorator


def _days_valid_pfr(url):
    # boxscores are static, but refresh quarterly to be sure
    if "boxscore" in url:
        return 90
    # important dates
    today = datetime.date.today()
    start_of_season = datetime.date(today.year, 8, 15)
    end_of_season = datetime.date(today.year, 2, 15)
    # check for a year in the filename
    m = re.search(r"(\d{4})", url)
    if m:
        # if it was a year prior to the current season, we're good
        year = int(m.group(1))
        cur_season = today.year - (today <= end_of_season)
        if year < cur_season:
            return 90
    # if it's the offseason, refresh cache twice a month
    if end_of_season < today < start_of_season:
        return 15
    # otherwise, refresh every 2 days
    return 2


def _days_valid_bkref(url):
    # boxscores are static, but refresh quarterly to be sure
    if "boxscore" in url:
        return 90
    # important dates
    today = datetime.date.today()
    start_of_season = datetime.date(today.year, 10, 1)
    end_of_season = datetime.date(today.year, 7, 1)
    # check for a year in the filename
    m = re.search(r"(\d{4})", url)
    if m:
        # if it was a year prior to the current season, we're good
        year = int(m.group(1))
        cur_season = today.year - (today <= end_of_season) + 1
        if year < cur_season:
            return 90
    # if it's the offseason, refresh cache once a month
    if end_of_season < today < start_of_season:
        return 30
    # otherwise, refresh every 2 days
    return 2


def _days_valid_cfb(url):
    # TODO: caching for CFB
    return 365


def cache(func):
    """Caches the HTML returned by the specified function `func`. Caches it in
    the user cache determined by the appdirs package.
    """

    CACHE_DIR = appdirs.user_cache_dir("sportsref", getpass.getuser())
    os.makedirs(CACHE_DIR, exist_ok=True)

    @functools.wraps(func)
    def wrapper(url):
        # hash based on the URL
        file_hash = hashlib.md5()
        encoded_url = url.encode(errors="replace")
        file_hash.update(encoded_url)
        file_hash = file_hash.hexdigest()
        filename = f"{CACHE_DIR}/{file_hash}"

        sport_id = None
        for a_base_url, a_sport_id in sportsref.SITE_ABBREV.items():
            if url.startswith(a_base_url):
                sport_id = a_sport_id
                break
        else:
            # TODO: log
            print(f"No sport ID found for {url}, not able to check cache")

        # check whether cache is valid or stale
        file_exists = os.path.isfile(filename)
        if sport_id and file_exists:
            cur_time = int(time.time())
            mod_time = int(os.path.getmtime(filename))
            days_since_mod = datetime.timedelta(seconds=(cur_time - mod_time)).days
            # TODO: refactor _days_valid_ functions to not use globals
            days_cache_valid = globals()[f"_days_valid_{sport_id}"](url)
            cache_is_valid = days_since_mod < days_cache_valid
        else:
            cache_is_valid = False

        # if file found and cache is valid, read from file
        allow_caching = sportsref.get_option("cache")
        if file_exists and cache_is_valid and allow_caching:
            with open(filename, "r", encoding="utf-8", errors="replace") as f:
                text = f.read()
        # otherwise, execute function and cache results
        else:
            text = func(url)
            with open(filename, "w+", encoding="utf-8") as f:
                f.write(text)
        return text

    return wrapper


def get_class_instance_key(cls, args, kwargs):
    """
    Returns a unique identifier for a class instantiation.
    """
    identifiers = [id(cls)]
    for arg in args:
        identifiers.append(id(arg))
    identifiers.extend((k, id(v)) for k, v in list(kwargs.items()))
    return tuple(sorted(identifiers))


# used as a metaclass for classes that should be memoized
# (technically not a decorator, but it's similar enough)
Cached = mementos.memento_factory("Cached", get_class_instance_key)


def memoize(fun):
    """A decorator for memoizing functions.

    Only works on functions that take simple arguments - arguments that take
    list-like or dict-like arguments will not be memoized, and this function
    will raise a TypeError.
    """

    @functools.wraps(fun)
    def wrapper(*args, **kwargs):

        do_memoization = sportsref.get_option("memoize")
        if not do_memoization:
            return fun(*args, **kwargs)

        hash_args = tuple(args)
        hash_kwargs = frozenset(sorted(kwargs.items()))
        key = (hash_args, hash_kwargs)

        def _copy(v):
            if isinstance(v, pq):
                return v.clone()
            else:
                return copy.deepcopy(v)

        try:
            ret = _copy(cache[key])
            return ret
        except KeyError:
            cache[key] = fun(*args, **kwargs)
            ret = _copy(cache[key])
            return ret
        except TypeError:
            print(
                f"memoization type error in function {fun.__name__} for arguments {key}"
            )
            raise

    cache = {}
    return wrapper


def kind_rpb(include_type=False):
    def decorator(fun):
        """Supports functions that return a DataFrame and have a `kind` keyword
        argument that specifies regular season ('R'), playoffs ('P'), or both
        ('B'). If given 'B', it will call the function with both 'R' and 'P'
        and concatenate the results.
        """

        @functools.wraps(fun)
        def wrapper(*args, **kwargs):
            kind = kwargs.get("kind", "R").upper()
            if kind == "B":
                kwargs["kind"] = "R"
                reg = fun(*args, **kwargs)
                if include_type:
                    reg["is_playoffs"] = False
                kwargs["kind"] = "P"
                poffs = fun(*args, **kwargs)
                if include_type:
                    poffs["is_playoffs"] = True
                return pd.concat((reg, poffs), ignore_index=True)
            else:
                df = fun(*args, **kwargs)
                if include_type:
                    df["is_playoffs"] = kind == "P"
                return df

        return wrapper

    return decorator



# flake8: noqa

SITE_ABBREV = {
    "http://www.pro-football-reference.com": "pfr",
    "http://www.basketball-reference.com": "bkref",
    "http://www.sports-reference.com/cfb": "ncaaf",
    "http://www.sports-reference.com/cbb": "ncaab",
}

from sportsref.options import get_option, set_option
from sportsref import decorators, utils, nfl, nba

__all__ = [
    "decorators",
    "utils",
    "nfl",
    "nba",
    "get_option",
    "set_option",
    "SITE_ABBREV",
]



import collections
import json
import os
import time

from pyquery import PyQuery as pq

from ... import decorators, utils
from .. import pbp

GPF_URL = "http://www.pro-football-reference.com/" "play-index/play_finder.cgi"

GPF_CONSTANTS_FILENAME = "GPFConstants.json"


def GamePlayFinder(**kwargs):
    """ Docstring will be filled in by __init__.py """

    querystring = _kwargs_to_qs(**kwargs)
    url = "{}?{}".format(GPF_URL, querystring)
    # if verbose, print url
    if kwargs.get("verbose", False):
        print(url)
    html = utils.get_html(url)
    doc = pq(html)

    # parse
    table = doc("table#all_plays")
    plays = utils.parse_table(table)

    # parse score column
    if "score" in plays.columns:
        oScore, dScore = list(zip(*plays.score.apply(lambda s: s.split("-"))))
        plays["teamScore"] = oScore
        plays["oppScore"] = dScore
    # add parsed pbp info
    if "description" in plays.columns:
        plays = pbp.expand_details(plays, detailCol="description")

    return plays


def _kwargs_to_qs(**kwargs):
    """Converts kwargs given to GPF to a querystring.

    :returns: the querystring.
    """
    # start with defaults
    inpOptDef = inputs_options_defaults()
    opts = {name: dct["value"] for name, dct in list(inpOptDef.items())}

    # clean up keys and values
    for k, v in list(kwargs.items()):
        # pID, playerID => player_id
        if k.lower() in ("pid", "playerid"):
            del kwargs[k]
            kwargs["player_id"] = v
        # player_id can accept rel URLs
        if k == "player_id":
            if v.startswith("/players/"):
                kwargs[k] = utils.rel_url_to_id(v)
        # bool => 'Y'|'N'
        if isinstance(v, bool):
            kwargs[k] = "Y" if v else "N"
        # tm, team => team_id
        if k.lower() in ("tm", "team"):
            del kwargs[k]
            kwargs["team_id"] = v
        # yr_min, yr_max => year_min, year_max
        if k.lower() in ("yr_min", "yr_max"):
            del kwargs[k]
            if k.lower() == "yr_min":
                kwargs["year_min"] = int(v)
            else:
                kwargs["year_max"] = int(v)
        # wk_min, wk_max => week_num_min, week_num_max
        if k.lower() in ("wk_min", "wk_max"):
            del kwargs[k]
            if k.lower() == "wk_min":
                kwargs["week_num_min"] = int(v)
            else:
                kwargs["week_num_max"] = int(v)
        # yr, year, yrs, years => year_min, year_max
        if k.lower() in ("yr", "year", "yrs", "years"):
            del kwargs[k]
            if isinstance(v, collections.Iterable):
                lst = list(v)
                kwargs["year_min"] = min(lst)
                kwargs["year_max"] = max(lst)
            elif isinstance(v, str):
                v = list(map(int, v.split(",")))
                kwargs["year_min"] = min(v)
                kwargs["year_max"] = max(v)
            else:
                kwargs["year_min"] = v
                kwargs["year_max"] = v
        # wk, week, wks, weeks => week_num_min, week_num_max
        if k.lower() in ("wk", "week", "wks", "weeks"):
            del kwargs[k]
            if isinstance(v, collections.Iterable):
                lst = list(v)
                kwargs["week_num_min"] = min(lst)
                kwargs["week_num_max"] = max(lst)
            elif isinstance(v, str):
                v = list(map(int, v.split(",")))
                kwargs["week_num_min"] = min(v)
                kwargs["week_num_max"] = max(v)
            else:
                kwargs["week_num_min"] = v
                kwargs["week_num_max"] = v
        # if playoff_round defined, then turn on playoff flag
        if k == "playoff_round":
            kwargs["game_type"] = "P"
        if isinstance(v, str):
            v = v.split(",")
        if not isinstance(v, collections.Iterable):
            v = [v]

    # reset values to blank for defined kwargs
    for k in kwargs:
        if k in opts:
            opts[k] = []

    # update based on kwargs
    for k, v in list(kwargs.items()):
        # if overwriting a default, overwrite it
        if k in opts:
            # if multiple values separated by commas, split em
            if isinstance(v, str):
                v = v.split(",")
            elif not isinstance(v, collections.Iterable):
                v = [v]
            for val in v:
                opts[k].append(val)

    opts["request"] = [1]

    qs = "&".join(
        "{}={}".format(name, val) for name, vals in sorted(opts.items()) for val in vals
    )

    return qs


@decorators.switch_to_dir(os.path.dirname(os.path.realpath(__file__)))
def inputs_options_defaults():
    """Handles scraping options for play finder form.

    :returns: {'name1': {'value': val, 'options': [opt1, ...] }, ... }

    """
    # set time variables
    if os.path.isfile(GPF_CONSTANTS_FILENAME):
        modtime = int(os.path.getmtime(GPF_CONSTANTS_FILENAME))
        curtime = int(time.time())
    # if file found and it's been <= a week
    if os.path.isfile(GPF_CONSTANTS_FILENAME) and curtime - modtime <= 7 * 24 * 60 * 60:

        # just read the dict from the cached file
        with open(GPF_CONSTANTS_FILENAME, "r") as const_f:
            def_dict = json.load(const_f)

    # otherwise, we must regenerate the dict and rewrite it
    else:

        print("Regenerating GPFConstants file")

        html = utils.get_html(GPF_URL)
        doc = pq(html)

        def_dict = {}
        # start with input elements
        for inp in doc("form#play_finder input[name]"):
            name = inp.attrib["name"]
            # add blank dict if not present
            if name not in def_dict:
                def_dict[name] = {"value": set(), "options": set(), "type": inp.type}

            val = inp.attrib.get("value", "")
            # handle checkboxes and radio buttons
            if inp.type in ("checkbox", "radio"):
                # deal with default value
                if "checked" in inp.attrib:
                    def_dict[name]["value"].add(val)
                # add to options
                def_dict[name]["options"].add(val)
            # handle other types of inputs (only other type is hidden?)
            else:
                def_dict[name]["value"].add(val)

        # for dropdowns (select elements)
        for sel in doc.items("form#play_finder select[name]"):
            name = sel.attr["name"]
            # add blank dict if not present
            if name not in def_dict:
                def_dict[name] = {"value": set(), "options": set(), "type": "select"}

            # deal with default value
            defaultOpt = sel("option[selected]")
            if len(defaultOpt):
                defaultOpt = defaultOpt[0]
                def_dict[name]["value"].add(defaultOpt.attrib.get("value", ""))
            else:
                def_dict[name]["value"].add(sel("option")[0].attrib.get("value", ""))

            # deal with options
            def_dict[name]["options"] = {
                opt.attrib["value"] for opt in sel("option") if opt.attrib.get("value")
            }

        # ignore QB kneels by default
        def_dict["include_kneels"]["value"] = ["0"]

        def_dict.pop("request", None)
        def_dict.pop("use_favorites", None)

        with open(GPF_CONSTANTS_FILENAME, "w+") as f:
            for k in def_dict:
                try:
                    def_dict[k]["value"] = sorted(list(def_dict[k]["value"]), key=int)
                    def_dict[k]["options"] = sorted(
                        list(def_dict[k]["options"]), key=int
                    )
                except Exception:
                    def_dict[k]["value"] = sorted(list(def_dict[k]["value"]))
                    def_dict[k]["options"] = sorted(list(def_dict[k]["options"]))
            json.dump(def_dict, f)

    return def_dict



from pyquery import PyQuery as pq

import sportsref


__all__ = ["Season"]


class Season(object, metaclass=sportsref.decorators.Cached):

    """Object representing a given NFL season."""

    def __init__(self, year):
        """Initializes a Season object for an NFL season.

        :year: The year of the season we want.
        """
        self.yr = int(year)

    def __eq__(self, other):
        return self.yr == other.yr

    def __hash__(self):
        return hash(self.yr)

    def __repr__(self):
        return "Season({})".format(self.yr)

    def _subpage_url(self, page):
        return sportsref.nfl.BASE_URL + "/years/{}/{}.htm".format(self.yr, page)

    @sportsref.decorators.memoize
    def get_main_doc(self):
        """Returns PyQuery object for the main season URL.
        :returns: PyQuery object.
        """
        url = sportsref.nfl.BASE_URL + "/years/{}/".format(self.yr)
        return pq(sportsref.utils.get_html(url))

    @sportsref.decorators.memoize
    def get_sub_doc(self, subpage):
        """Returns PyQuery object for a given subpage URL.
        :subpage: The subpage of the season, e.g. 'per_game'.
        :returns: PyQuery object.
        """
        html = sportsref.utils.get_html(self._subpage_url(subpage))
        return pq(html)

    @sportsref.decorators.memoize
    def get_team_ids(self):
        """Returns a list of the team IDs for the given year.
        :returns: List of team IDs.
        """
        return sportsref.nfl.teams.list_teams(self.yr)

    @sportsref.decorators.memoize
    def team_ids_to_names(self):
        """Mapping from 3-letter team IDs to full team names.

        :returns: Dictionary with team IDs as keys and full team strings as
            values.
        """
        return sportsref.nfl.teams.team_names(self.yr)

    @sportsref.decorators.memoize
    def team_names_to_ids(self):
        """Mapping from full team names to 3-letter team IDs.
        :returns: Dictionary with tean names as keys and team IDs as values.
        """
        return sportsref.nfl.teams.team_ids(self.yr)

    @sportsref.decorators.memoize
    def _get_player_stats_table(self, subpage, table_id):
        """Helper function for player season stats.

        :identifier: string identifying the type of stat, e.g. 'passing'.
        :returns: A DataFrame of stats.
        """
        doc = self.get_sub_doc(subpage)
        table = doc("table#{}".format(table_id))
        df = sportsref.utils.parse_table(table)
        return df

    def player_stats_passing(self):
        """Returns a DataFrame of passing player stats for a season."""
        return self._get_player_stats_table("passing", "passing")

    def player_stats_rushing(self):
        """Returns a DataFrame of rushing player stats for a season."""
        return self._get_player_stats_table("rushing", "rushing_and_receiving")

    def player_stats_receiving(self):
        """Returns a DataFrame of receiving player stats for a season."""
        return self._get_player_stats_table("receiving", "receiving")



from . import finders
from . import teams
from . import players
from . import boxscores

# from . import winProb
from . import pbp

from .players import Player
from .seasons import Season
from .teams import Team
from .boxscores import BoxScore
from .finders import GamePlayFinder, PlayerSeasonFinder

BASE_URL = "http://www.pro-football-reference.com"

# modules/variables to expose
__all__ = [
    "BASE_URL",
    "finders",
    "GamePlayFinder",
    "PlayerSeasonFinder",
    "boxscores",
    "BoxScore",
    "players",
    "Player",
    "seasons",
    "Season",
    "teams",
    "Team",
    # "winProb",
    "pbp",
]



import collections
import json
import os
import time
import urllib.parse

from pyquery import PyQuery as pq

from ... import decorators, utils

PSF_URL = "http://www.pro-football-reference.com/" "play-index/psl_finder.cgi"

PSF_CONSTANTS_FILENAME = "PSFConstants.json"


def PlayerSeasonFinder(**kwargs):
    """ Docstring will be filled in by __init__.py """

    if "offset" not in kwargs:
        kwargs["offset"] = 0

    playerSeasons = []
    while True:
        querystring = _kwargs_to_qs(**kwargs)
        url = "{}?{}".format(PSF_URL, querystring)
        if kwargs.get("verbose", False):
            print(url)
        html = utils.get_html(url)
        doc = pq(html)
        table = doc("table#results")
        df = utils.parse_table(table)
        if df.empty:
            break

        thisSeason = list(zip(df.player_id, df.year))
        playerSeasons.extend(thisSeason)

        if doc('*:contains("Next Page")'):
            kwargs["offset"] += 100
        else:
            break

    return playerSeasons


def _kwargs_to_qs(**kwargs):
    """Converts kwargs given to PSF to a querystring.

    :returns: the querystring.
    """
    # start with defaults
    inpOptDef = inputs_options_defaults()
    opts = {name: dct["value"] for name, dct in list(inpOptDef.items())}

    # clean up keys and values
    for k, v in list(kwargs.items()):
        del kwargs[k]
        # bool => 'Y'|'N'
        if isinstance(v, bool):
            kwargs[k] = "Y" if v else "N"
        # tm, team => team_id
        elif k.lower() in ("tm", "team"):
            kwargs["team_id"] = v
        # yr, year, yrs, years => year_min, year_max
        elif k.lower() in ("yr", "year", "yrs", "years"):
            if isinstance(v, collections.Iterable):
                lst = list(v)
                kwargs["year_min"] = min(lst)
                kwargs["year_max"] = max(lst)
            elif isinstance(v, str):
                v = list(map(int, v.split(",")))
                kwargs["year_min"] = min(v)
                kwargs["year_max"] = max(v)
            else:
                kwargs["year_min"] = v
                kwargs["year_max"] = v
        # pos, position, positions => pos[]
        elif k.lower() in ("pos", "position", "positions"):
            if isinstance(v, str):
                v = v.split(",")
            elif not isinstance(v, collections.Iterable):
                v = [v]
            kwargs["pos[]"] = v
        # draft_pos, ... => draft_pos[]
        elif k.lower() in (
            "draft_pos",
            "draftpos",
            "draftposition",
            "draftpositions",
            "draft_position",
            "draft_positions",
        ):
            if isinstance(v, str):
                v = v.split(",")
            elif not isinstance(v, collections.Iterable):
                v = [v]
            kwargs["draft_pos[]"] = v
        # if not one of these cases, put it back in kwargs
        else:
            kwargs[k] = v

    # update based on kwargs
    for k, v in list(kwargs.items()):
        # if overwriting a default, overwrite it (with a list so the
        # opts -> querystring list comp works)
        if k in opts or k in ("pos[]", "draft_pos[]"):
            # if multiple values separated by commas, split em
            if isinstance(v, str):
                v = v.split(",")
            # otherwise, make sure it's a list
            elif not isinstance(v, collections.Iterable):
                v = [v]
            # then, add list of values to the querystring dict *opts*
            opts[k] = v
        if "draft" in k:
            opts["draft"] = [1]

    opts["request"] = [1]
    opts["offset"] = [kwargs.get("offset", 0)]

    qs = "&".join(
        "{}={}".format(urllib.parse.quote_plus(name), val)
        for name, vals in sorted(opts.items())
        for val in vals
    )

    return qs


@decorators.switch_to_dir(os.path.dirname(os.path.realpath(__file__)))
def inputs_options_defaults():
    """Handles scraping options for player-season finder form.

    :returns: {'name1': {'value': val, 'options': [opt1, ...] }, ... }
    """
    # set time variables
    if os.path.isfile(PSF_CONSTANTS_FILENAME):
        modtime = int(os.path.getmtime(PSF_CONSTANTS_FILENAME))
        curtime = int(time.time())
    # if file found and it's been <= a week
    if os.path.isfile(PSF_CONSTANTS_FILENAME) and curtime - modtime <= 7 * 24 * 60 * 60:

        # just read the dict from cached file
        with open(PSF_CONSTANTS_FILENAME, "r") as const_f:
            def_dict = json.load(const_f)

    # otherwise, we must regenerate the dict and rewrite it
    else:

        print("Regenerating PSFConstants file")

        html = utils.get_html(PSF_URL)
        doc = pq(html)

        def_dict = {}
        # start with input elements
        for inp in doc("form#psl_finder input[name]"):
            name = inp.attrib["name"]
            # add blank dict if not present
            if name not in def_dict:
                def_dict[name] = {
                    "value": set(),
                    "options": set(),
                    "type": inp.attrib["type"],
                }

            # handle checkboxes and radio buttons
            if inp.attrib["type"] in ("checkbox", "radio"):
                # deal with default value
                if "checked" in inp.attrib:
                    def_dict[name]["value"].add(inp.attrib["value"])
                # add to options
                def_dict[name]["options"].add(inp.attrib["value"])
            # handle other types of inputs (only other type is hidden?)
            else:
                def_dict[name]["value"].add(inp.attrib.get("value", ""))

        # deal with dropdowns (select elements)
        for sel in doc.items("form#psl_finder select[name]"):
            name = sel.attr["name"]
            # add blank dict if not present
            if name not in def_dict:
                def_dict[name] = {"value": set(), "options": set(), "type": "select"}

            # deal with default value
            defaultOpt = sel("option[selected]")
            if len(defaultOpt):
                defaultOpt = defaultOpt[0]
                def_dict[name]["value"].add(defaultOpt.attrib.get("value", ""))
            else:
                def_dict[name]["value"].add(sel("option")[0].attrib.get("value", ""))

            # deal with options
            def_dict[name]["options"] = {
                opt.attrib["value"] for opt in sel("option") if opt.attrib.get("value")
            }

        def_dict.pop("request", None)
        def_dict.pop("use_favorites", None)

        with open(PSF_CONSTANTS_FILENAME, "w+") as f:
            for k in def_dict:
                try:
                    def_dict[k]["value"] = sorted(list(def_dict[k]["value"]), key=int)
                    def_dict[k]["options"] = sorted(
                        list(def_dict[k]["options"]), key=int
                    )
                except Exception:
                    def_dict[k]["value"] = sorted(list(def_dict[k]["value"]))
                    def_dict[k]["options"] = sorted(list(def_dict[k]["options"]))
            json.dump(def_dict, f)

    return def_dict



import ctypes
import multiprocessing
import re
import threading
import time

import pandas as pd
import requests
from pyquery import PyQuery as pq

import sportsref

# time between requests, in seconds
THROTTLE_DELAY = 0.5

# variables used to throttle requests across processes
throttle_thread_lock = threading.Lock()
throttle_process_lock = multiprocessing.Lock()
last_request_time = multiprocessing.Value(
    ctypes.c_longdouble, time.time() - 10 * THROTTLE_DELAY
)


@sportsref.decorators.cache
def get_html(url):
    """Gets the HTML for the given URL using a GET request.

    :url: the absolute URL of the desired page.
    :returns: a string of HTML.
    """
    global last_request_time
    with throttle_process_lock:
        with throttle_thread_lock:
            # sleep until THROTTLE_DELAY secs have passed since last request
            wait_left = THROTTLE_DELAY - (time.time() - last_request_time.value)
            if wait_left > 0:
                time.sleep(wait_left)

            # make request
            response = requests.get(url)

            # update last request time for throttling
            last_request_time.value = time.time()

    # raise ValueError on 4xx status code, get rid of comments, and return
    if 400 <= response.status_code < 500:
        raise ValueError(
            f'Status Code {response.status_code} received fetching URL "{url}"'
        )
    html = response.text
    html = html.replace("<!--", "").replace("-->", "")

    return html


def parse_table(table, flatten=True, footer=False):
    """Parses a table from sports-reference sites into a pandas dataframe.

    :param table: the PyQuery object representing the HTML table
    :param flatten: if True, flattens relative URLs to IDs. otherwise, leaves
        all fields as text without cleaning.
    :param footer: If True, returns the summary/footer of the page. Recommended
        to use this with flatten=False. Defaults to False.
    :returns: pd.DataFrame
    """
    if not len(table):
        return pd.DataFrame()

    # get columns
    columns = [
        c.attrib["data-stat"] for c in table("thead tr:not([class]) th[data-stat]")
    ]

    # get data
    rows = list(
        table("tbody tr" if not footer else "tfoot tr")
        .not_(".thead, .stat_total, .stat_average")
        .items()
    )
    data = [
        [flatten_links(td) if flatten else td.text() for td in row.items("th,td")]
        for row in rows
    ]

    # make DataFrame
    df = pd.DataFrame(data, columns=columns, dtype="float")

    # add has_class columns
    allClasses = set(
        cls for row in rows if row.attr["class"] for cls in row.attr["class"].split()
    )
    for cls in allClasses:
        df["has_class_" + cls] = [
            bool(row.attr["class"] and cls in row.attr["class"].split()) for row in rows
        ]

    # cleaning the DataFrame

    df.drop(["ranker", "Xxx", "Yyy", "Zzz"], axis=1, inplace=True, errors="ignore")

    # year_id -> year (as int)
    if "year_id" in df.columns:
        df.rename(columns={"year_id": "year"}, inplace=True)
        if flatten:
            df.year = df.year.fillna(method="ffill")
            df["year"] = df.year.map(lambda s: str(s)[:4]).astype(int)

    # pos -> position
    if "pos" in df.columns:
        df.rename(columns={"pos": "position"}, inplace=True)

    # boxscore_word, game_date -> boxscore_id and separate into Y, M, D columns
    for bs_id_col in ("boxscore_word", "game_date", "box_score_text"):
        if bs_id_col in df.columns:
            df.rename(columns={bs_id_col: "boxscore_id"}, inplace=True)
            break

    # ignore *, +, and other characters used to note things
    df.replace(re.compile(r"[\*\+\u2605]", re.U), "", inplace=True)
    for col in df.columns:
        if hasattr(df[col], "str"):
            df[col] = df[col].str.strip()

    # player -> player_id and/or player_name
    if "player" in df.columns:
        if flatten:
            df.rename(columns={"player": "player_id"}, inplace=True)
            # when flattening, keep a column for names
            player_names = parse_table(table, flatten=False)["player_name"]
            df["player_name"] = player_names
        else:
            df.rename(columns={"player": "player_name"}, inplace=True)

    # team, team_name -> team_id
    for team_col in ("team", "team_name"):
        if team_col in df.columns:
            # first, get rid of faulty rows
            df = df.loc[~df[team_col].isin(["XXX"])]
            if flatten:
                df.rename(columns={team_col: "team_id"}, inplace=True)

    # season -> int
    if "season" in df.columns and flatten:
        df["season"] = df["season"].astype(int)

    # handle date_game columns (different types)
    if "date_game" in df.columns and flatten:
        date_re = r"month=(?P<month>\d+)&day=(?P<day>\d+)&year=(?P<year>\d+)"
        date_df = df["date_game"].str.extract(date_re, expand=True)
        if date_df.notnull().all(axis=1).any():
            df = pd.concat((df, date_df), axis=1)
        else:
            df.rename(columns={"date_game": "boxscore_id"}, inplace=True)

    # game_location -> is_home
    if "game_location" in df.columns and flatten:
        df["game_location"] = df["game_location"].isnull()
        df.rename(columns={"game_location": "is_home"}, inplace=True)

    # mp: (min:sec) -> float(min + sec / 60), notes -> NaN, new column
    if "mp" in df.columns and df.dtypes["mp"] == object and flatten:
        mp_df = (
            df["mp"].str.extract(r"(?P<m>\d+):(?P<s>\d+)", expand=True).astype(float)
        )
        no_match = mp_df.isnull().all(axis=1)
        if no_match.any():
            df.loc[no_match, "note"] = df.loc[no_match, "mp"]
        df["mp"] = mp_df["m"] + mp_df["s"] / 60

    # converts number-y things to floats
    def convert_to_float(val):
        # percentages: (number%) -> float(number * 0.01)
        m = re.search(r"([-\.\d]+)\%", val if isinstance(val, str) else str(val), re.U)
        try:
            if m:
                return float(m.group(1)) / 100 if m else val
            if m:
                return int(m.group(1)) + int(m.group(2)) / 60
        except ValueError:
            return val
        # salaries: $ABC,DEF,GHI -> float(ABCDEFGHI)
        m = re.search(r"\$[\d,]+", val if isinstance(val, str) else str(val), re.U)
        try:
            if m:
                return float(re.sub(r"\$|,", "", val))
        except Exception:
            return val
        # generally try to coerce to float, unless it's an int or bool
        try:
            if isinstance(val, (int, bool)):
                return val
            else:
                return float(val)
        except Exception:
            return val

    if flatten:
        df = df.applymap(convert_to_float)

    df = df.loc[df.astype(bool).any(axis=1)]

    return df


def parse_info_table(table):
    """Parses an info table, like the "Game Info" table or the "Officials"
    table on the PFR Boxscore page. Keys are lower case and have spaces/special
    characters converted to underscores.

    :table: PyQuery object representing the HTML table.
    :returns: A dictionary representing the information.
    """
    ret = {}
    for tr in list(table("tr").not_(".thead").items()):
        th, td = list(tr("th, td").items())
        key = th.text().lower()
        key = re.sub(r"\W", "_", key)
        val = sportsref.utils.flatten_links(td)
        ret[key] = val
    return ret


def parse_awards_table(table):
    """Parses an awards table, like the "Pro Bowls" table on a PFR player page.

    :table: PyQuery object representing the HTML table.
    :returns: A list of the entries in the table, with flattened links.
    """
    return [flatten_links(tr) for tr in list(table("tr").items())]


def flatten_links(td, _recurse=False):
    """Flattens relative URLs within text of a table cell to IDs and returns
    the result.

    :td: the PyQuery object for the HTML to convert
    :returns: the string with the links flattened to IDs
    """

    # helper function to flatten individual strings/links
    def _flatten_node(c):
        if isinstance(c, str):
            return c
        elif "href" in c.attrib:
            c_id = rel_url_to_id(c.attrib["href"])
            return c_id if c_id else c.text_content()
        else:
            return flatten_links(pq(c), _recurse=True)

    # if there's no text, just return None
    if td is None or not td.text():
        return "" if _recurse else None

    td.remove("span.note")
    return "".join(_flatten_node(c) for c in td.contents())


@sportsref.decorators.memoize
def rel_url_to_id(url):
    """Converts a relative URL to a unique ID.

    Here, 'ID' refers generally to the unique ID for a given 'type' that a
    given datum has. For example, 'BradTo00' is Tom Brady's player ID - this
    corresponds to his relative URL, '/players/B/BradTo00.htm'. Similarly,
    '201409070dal' refers to the boxscore of the SF @ DAL game on 09/07/14.

    Supported types:
    * player/...
    * boxscores/...
    * teams/...
    * years/...
    * leagues/...
    * awards/...
    * coaches/...
    * officials/...
    * schools/...
    * schools/high_schools.cgi?id=...

    :returns: ID associated with the given relative URL.
    """
    yearRegex = r".*/years/(\d{4}).*|.*/gamelog/(\d{4}).*"
    playerRegex = r".*/players/(?:\w/)?(.+?)(?:/|\.html?)"
    boxscoresRegex = r".*/boxscores/(.+?)\.html?"
    teamRegex = r".*/teams/(\w{3})/.*"
    coachRegex = r".*/coaches/(.+?)\.html?"
    stadiumRegex = r".*/stadiums/(.+?)\.html?"
    refRegex = r".*/officials/(.+?r)\.html?"
    collegeRegex = r".*/schools/(\S+?)/.*|.*college=([^&]+)"
    hsRegex = r".*/schools/high_schools\.cgi\?id=([^\&]{8})"
    bsDateRegex = r".*/boxscores/index\.f?cgi\?(month=\d+&day=\d+&year=\d+)"
    leagueRegex = r".*/leagues/(.*_\d{4}).*"
    awardRegex = r".*/awards/(.+)\.htm"

    regexes = [
        yearRegex,
        playerRegex,
        boxscoresRegex,
        teamRegex,
        coachRegex,
        stadiumRegex,
        refRegex,
        collegeRegex,
        hsRegex,
        bsDateRegex,
        leagueRegex,
        awardRegex,
    ]

    for regex in regexes:
        match = re.match(regex, url, re.I)
        if match:
            return [_f for _f in match.groups() if _f][0]

    # things we don't want to match but don't want to print a WARNING
    if any(url.startswith(s) for s in ("/play-index/",)):
        return url

    print(f'WARNING. NO MATCH WAS FOUND FOR "{url}"')
    return url



OPTIONS = {"cache": True, "memoize": True}


def get_option(option):
    option = option.lower()
    if option in OPTIONS:
        return OPTIONS[option]
    else:
        # TODO: log
        print(f"option {option} not recognized")
        return None


def set_option(option, value):
    option = option.lower()
    if option in OPTIONS:
        OPTIONS[option] = value
    else:
        # TODO: log
        print("option {option} not recognized")



import pandas as pd
from pyquery import PyQuery as pq

import sportsref


class Season(object, metaclass=sportsref.decorators.Cached):

    """Object representing a given NBA season."""

    def __init__(self, year):
        """Initializes a Season object for an NBA season.

        :year: The year of the season we want.
        """
        self.yr = int(year)

    def __eq__(self, other):
        return self.yr == other.yr

    def __hash__(self):
        return hash(self.yr)

    def __repr__(self):
        return "Season({})".format(self.yr)

    def _subpage_url(self, page):
        return sportsref.nba.BASE_URL + "/leagues/NBA_{}_{}.html".format(self.yr, page)

    @sportsref.decorators.memoize
    def get_main_doc(self):
        """Returns PyQuery object for the main season URL.
        :returns: PyQuery object.
        """
        url = sportsref.nba.BASE_URL + "/leagues/NBA_{}.html".format(self.yr)
        return pq(sportsref.utils.get_html(url))

    @sportsref.decorators.memoize
    def get_sub_doc(self, subpage):
        """Returns PyQuery object for a given subpage URL.
        :subpage: The subpage of the season, e.g. 'per_game'.
        :returns: PyQuery object.
        """
        html = sportsref.utils.get_html(self._subpage_url(subpage))
        return pq(html)

    @sportsref.decorators.memoize
    def get_team_ids(self):
        """Returns a list of the team IDs for the given year.
        :returns: List of team IDs.
        """
        df = self.team_stats_per_game()
        if not df.empty:
            return df.index.tolist()
        else:
            print("ERROR: no teams found")
            return []

    @sportsref.decorators.memoize
    def team_ids_to_names(self):
        """Mapping from 3-letter team IDs to full team names.
        :returns: Dictionary with team IDs as keys and full team strings as
        values.
        """
        doc = self.get_main_doc()
        table = doc("table#team-stats-per_game")
        flattened = sportsref.utils.parse_table(table, flatten=True)
        unflattened = sportsref.utils.parse_table(table, flatten=False)
        team_ids = flattened["team_id"]
        team_names = unflattened["team_name"]
        if len(team_names) != len(team_ids):
            raise Exception("team names and team IDs don't align")
        return dict(list(zip(team_ids, team_names)))

    @sportsref.decorators.memoize
    def team_names_to_ids(self):
        """Mapping from full team names to 3-letter team IDs.
        :returns: Dictionary with tean names as keys and team IDs as values.
        """
        d = self.team_ids_to_names()
        return {v: k for k, v in list(d.items())}

    @sportsref.decorators.memoize
    @sportsref.decorators.kind_rpb(include_type=True)
    def schedule(self, kind="R"):
        """Returns a list of BoxScore IDs for every game in the season.
        Only needs to handle 'R' or 'P' options because decorator handles 'B'.

        :param kind: 'R' for regular season, 'P' for playoffs, 'B' for both.
            Defaults to 'R'.
        :returns: DataFrame of schedule information.
        :rtype: pd.DataFrame
        """
        kind = kind.upper()[0]
        dfs = []

        # get games from each month
        for month in (
            "october",
            "november",
            "december",
            "january",
            "february",
            "march",
            "april",
            "may",
            "june",
        ):
            try:
                doc = self.get_sub_doc("games-{}".format(month))
            except ValueError:
                continue
            table = doc("table#schedule")
            df = sportsref.utils.parse_table(table)
            dfs.append(df)
        df = pd.concat(dfs).reset_index(drop=True)

        # figure out how many regular season games
        try:
            sportsref.utils.get_html(
                "{}/playoffs/NBA_{}.html".format(sportsref.nba.BASE_URL, self.yr)
            )
            is_past_season = True
        except ValueError:
            is_past_season = False

        if is_past_season:
            team_per_game = self.team_stats_per_game()
            n_reg_games = int(team_per_game.g.sum() // 2)
        else:
            n_reg_games = len(df)

        # subset appropriately based on `kind`
        if kind == "P":
            return df.iloc[n_reg_games:]
        else:
            return df.iloc[:n_reg_games]

    def finals_winner(self):
        """Returns the team ID for the winner of that year's NBA Finals.
        :returns: 3-letter team ID for champ.
        """
        raise NotImplementedError("nba.Season.finals_winner")

    def finals_loser(self):
        """Returns the team ID for the loser of that year's NBA Finals.
        :returns: 3-letter team ID for runner-up.
        """
        raise NotImplementedError("nba.Season.finals_loser")

    def standings(self):
        """Returns a DataFrame containing standings information."""
        doc = self.get_sub_doc("standings")

        east_table = doc("table#confs_standings_E")
        east_df = sportsref.utils.parse_table(east_table)
        east_df.sort_values("wins", ascending=False, inplace=True)
        east_df["seed"] = list(range(1, len(east_df) + 1))
        east_df["conference"] = "E"

        west_table = doc("table#confs_standings_W")
        west_df = sportsref.utils.parse_table(west_table)
        west_df.sort_values("wins", ascending=False, inplace=True)
        west_df["seed"] = list(range(1, len(west_df) + 1))
        west_df["conference"] = "W"

        full_df = pd.concat([east_df, west_df], axis=0).reset_index(drop=True)
        full_df["gb"] = [
            gb if isinstance(gb, int) or isinstance(gb, float) else 0
            for gb in full_df["gb"]
        ]
        full_df = full_df.drop("has_class_full_table", axis=1)

        expanded_table = doc("table#expanded_standings")
        expanded_df = sportsref.utils.parse_table(expanded_table)

        full_df = pd.merge(full_df, expanded_df, on="team_id")
        return full_df

    @sportsref.decorators.memoize
    def _get_team_stats_table(self, selector):
        """Helper function for stats tables on season pages. Returns a
        DataFrame."""
        doc = self.get_main_doc()
        table = doc(selector)
        df = sportsref.utils.parse_table(table)
        df.set_index("team_id", inplace=True)
        return df

    def team_stats_per_game(self):
        """Returns a Pandas DataFrame of each team's basic per-game stats for
        the season."""
        return self._get_team_stats_table("table#team-stats-per_game")

    def opp_stats_per_game(self):
        """Returns a Pandas DataFrame of each team's opponent's basic per-game
        stats for the season."""
        return self._get_team_stats_table("table#opponent-stats-per_game")

    def team_stats_totals(self):
        """Returns a Pandas DataFrame of each team's basic stat totals for the
        season."""
        return self._get_team_stats_table("table#team-stats-base")

    def opp_stats_totals(self):
        """Returns a Pandas DataFrame of each team's opponent's basic stat
        totals for the season."""
        return self._get_team_stats_table("table#opponent-stats-base")

    def misc_stats(self):
        """Returns a Pandas DataFrame of miscellaneous stats about each team's
        season."""
        return self._get_team_stats_table("table#misc_stats")

    def team_stats_shooting(self):
        """Returns a Pandas DataFrame of each team's shooting stats for the
        season."""
        return self._get_team_stats_table("table#team_shooting")

    def opp_stats_shooting(self):
        """Returns a Pandas DataFrame of each team's opponent's shooting stats
        for the season."""
        return self._get_team_stats_table("table#opponent_shooting")

    @sportsref.decorators.memoize
    def _get_player_stats_table(self, identifier):
        """Helper function for player season stats.

        :identifier: string identifying the type of stat, e.g. 'per_game'.
        :returns: A DataFrame of stats.
        """
        doc = self.get_sub_doc(identifier)
        table = doc("table#{}_stats".format(identifier))
        df = sportsref.utils.parse_table(table)
        return df

    def player_stats_per_game(self):
        """Returns a DataFrame of per-game player stats for a season."""
        return self._get_player_stats_table("per_game")

    def player_stats_totals(self):
        """Returns a DataFrame of player stat totals for a season."""
        return self._get_player_stats_table("totals")

    def player_stats_per36(self):
        """Returns a DataFrame of player per-36 min stats for a season."""
        return self._get_player_stats_table("per_minute")

    def player_stats_per100(self):
        """Returns a DataFrame of player per-100 poss stats for a season."""
        return self._get_player_stats_table("per_poss")

    def player_stats_advanced(self):
        """Returns a DataFrame of player per-100 poss stats for a season."""
        return self._get_player_stats_table("advanced")

    def mvp_voting(self):
        """Returns a DataFrame containing information about MVP voting."""
        raise NotImplementedError("nba.Season.mvp_voting")

    def roy_voting(self):
        """Returns a DataFrame containing information about ROY voting."""
        url = "{}/awards/awards_{}.html".format(sportsref.nba.BASE_URL, self.yr)
        doc = pq(sportsref.utils.get_html(url))
        table = doc("table#roy")
        df = sportsref.utils.parse_table(table)
        return df



import re

import numpy as np
import pandas as pd
from pyquery import PyQuery as pq

import sportsref

__all__ = ["team_names", "team_ids", "list_teams", "Team"]


@sportsref.decorators.memoize
def team_names(year):
    """Returns a mapping from team ID to full team name for a given season.
    Example of a full team name: "New England Patriots"

    :year: The year of the season in question (as an int).
    :returns: A dictionary with teamID keys and full team name values.
    """
    doc = pq(sportsref.utils.get_html(sportsref.nfl.BASE_URL + "/teams/"))
    active_table = doc("table#teams_active")
    active_df = sportsref.utils.parse_table(active_table)
    inactive_table = doc("table#teams_inactive")
    inactive_df = sportsref.utils.parse_table(inactive_table)
    df = pd.concat((active_df, inactive_df))
    df = df.loc[~df["has_class_partial_table"]]
    ids = df.team_id.str[:3].values
    names = [tr("th a") for tr in list(active_table("tr").items())]
    names.extend(tr("th a") for tr in list(inactive_table("tr").items()))
    names = [_f for _f in names if _f]
    names = [lst[0].text_content() for lst in names]
    # combine IDs and team names into pandas series
    series = pd.Series(names, index=ids)
    # create a mask to filter to teams from the given year
    mask = ((df.year_min <= year) & (year <= df.year_max)).values
    # filter, convert to a dict, and return
    return series[mask].to_dict()


@sportsref.decorators.memoize
def team_ids(year):
    """Returns a mapping from team name to team ID for a given season. Inverse
    mapping of team_names. Example of a full team name: "New England Patriots"

    :year: The year of the season in question (as an int).
    :returns: A dictionary with full team name keys and teamID values.
    """
    names = team_names(year)
    return {v: k for k, v in list(names.items())}


@sportsref.decorators.memoize
def list_teams(year):
    """Returns a list of team IDs for a given season.

    :year: The year of the season in question (as an int).
    :returns: A list of team IDs.
    """
    return list(team_names(year).keys())


class Team(object, metaclass=sportsref.decorators.Cached):
    def __init__(self, teamID):
        self.teamID = teamID

    def __eq__(self, other):
        return self.teamID == other.teamID

    def __hash__(self):
        return hash(self.teamID)

    def __repr__(self):
        return "Team({})".format(self.teamID)

    def __str__(self):
        return self.name()

    def __reduce__(self):
        return Team, (self.teamID,)

    @sportsref.decorators.memoize
    def team_year_url(self, yr_str):
        return sportsref.nfl.BASE_URL + "/teams/{}/{}.htm".format(self.teamID, yr_str)

    @sportsref.decorators.memoize
    def get_main_doc(self):
        relURL = "/teams/{}".format(self.teamID)
        teamURL = sportsref.nfl.BASE_URL + relURL
        mainDoc = pq(sportsref.utils.get_html(teamURL))
        return mainDoc

    @sportsref.decorators.memoize
    def get_year_doc(self, yr_str):
        return pq(sportsref.utils.get_html(self.team_year_url(yr_str)))

    @sportsref.decorators.memoize
    def name(self):
        """Returns the real name of the franchise given the team ID.

        Examples:
        'nwe' -> 'New England Patriots'
        'sea' -> 'Seattle Seahawks'

        :returns: A string corresponding to the team's full name.
        """
        doc = self.get_main_doc()
        headerwords = doc("div#meta h1")[0].text_content().split()
        lastIdx = headerwords.index("Franchise")
        teamwords = headerwords[:lastIdx]
        return " ".join(teamwords)

    @sportsref.decorators.memoize
    def roster(self, year):
        """Returns the roster table for the given year.

        :year: The year for which we want the roster; defaults to current year.
        :returns: A DataFrame containing roster information for that year.
        """
        doc = self.get_year_doc("{}_roster".format(year))
        roster_table = doc("table#games_played_team")
        df = sportsref.utils.parse_table(roster_table)
        starter_table = doc("table#starters")
        if not starter_table.empty:
            start_df = sportsref.utils.parse_table(starter_table)
            start_df = start_df.dropna(axis=0, subset=["position"])
            starters = start_df.set_index("position").player_id
            df["is_starter"] = df.player_id.isin(starters)
            df["starting_pos"] = df.player_id.map(
                lambda pid: (
                    starters[starters == pid].index[0]
                    if pid in starters.values
                    else None
                )
            )
        return df

    @sportsref.decorators.memoize
    def boxscores(self, year):
        """Gets list of BoxScore objects corresponding to the box scores from
        that year.

        :year: The year for which we want the boxscores; defaults to current
        year.
        :returns: np.array of strings representing boxscore IDs.
        """
        doc = self.get_year_doc(year)
        table = doc("table#games")
        df = sportsref.utils.parse_table(table)
        if df.empty:
            return np.array([])
        return df.boxscore_id.values

    @sportsref.decorators.memoize
    def _year_info_pq(self, year, keyword):
        """Returns a PyQuery object containing the info from the meta div at
        the top of the team year page with the given keyword.

        :year: Int representing the season.
        :keyword: A keyword to filter to a single p tag in the meta div.
        :returns: A PyQuery object for the selected p element.
        """
        doc = self.get_year_doc(year)
        p_tags = doc("div#meta div:not(.logo) p")
        texts = [p_tag.text_content().strip() for p_tag in p_tags]
        try:
            return next(
                pq(p_tag)
                for p_tag, text in zip(p_tags, texts)
                if keyword.lower() in text.lower()
            )
        except StopIteration:
            if len(texts):
                raise ValueError("Keyword not found in any p tag.")
            else:
                raise ValueError("No meta div p tags found.")

    # TODO: add functions for OC, DC, PF, PA, W-L, etc.
    # TODO: Also give a function at BoxScore.homeCoach and BoxScore.awayCoach
    # TODO: BoxScore needs a gameNum function to do this?

    @sportsref.decorators.memoize
    def head_coaches_by_game(self, year):
        """Returns head coach data by game.

        :year: An int representing the season in question.
        :returns: An array with an entry per game of the season that the team
        played (including playoffs). Each entry is the head coach's ID for that
        game in the season.
        """
        coach_str = self._year_info_pq(year, "Coach").text()
        regex = r"(\S+?) \((\d+)-(\d+)-(\d+)\)"
        coachAndTenure = []
        m = True
        while m:
            m = re.search(regex, coach_str)
            coachID, wins, losses, ties = m.groups()
            # nextIndex = m.end(4) + 1
            # coachStr = coachStr[nextIndex:]
            tenure = int(wins) + int(losses) + int(ties)
            coachAndTenure.append((coachID, tenure))

        coachIDs = [cID for cID, games in coachAndTenure for _ in range(games)]
        return np.array(coachIDs[::-1])

    @sportsref.decorators.memoize
    def wins(self, year):
        """Returns the # of regular season wins a team in a year.

        :year: The year for the season in question.
        :returns: The number of regular season wins.
        """
        schedule = self.schedule(year)
        if schedule.empty:
            return np.nan
        return schedule.query("week_num <= 17").is_win.sum()

    @sportsref.decorators.memoize
    def schedule(self, year):
        """Returns a DataFrame with schedule information for the given year.

        :year: The year for the season in question.
        :returns: Pandas DataFrame with schedule information.
        """
        doc = self.get_year_doc(year)
        table = doc("table#games")
        df = sportsref.utils.parse_table(table)
        if df.empty:
            return pd.DataFrame()
        df = df.loc[df["week_num"].notnull()]
        df["week_num"] = np.arange(len(df)) + 1
        df["is_win"] = df["game_outcome"] == "W"
        df["is_loss"] = df["game_outcome"] == "L"
        df["is_tie"] = df["game_outcome"] == "T"
        df["is_bye"] = df["game_outcome"].isnull()
        df["is_ot"] = df["overtime"].notnull()
        return df

    @sportsref.decorators.memoize
    def srs(self, year):
        """Returns the SRS (Simple Rating System) for a team in a year.

        :year: The year for the season in question.
        :returns: A float of SRS.
        """
        try:
            srs_text = self._year_info_pq(year, "SRS").text()
        except ValueError:
            return None
        m = re.match(r"SRS\s*?:\s*?(\S+)", srs_text)
        if m:
            return float(m.group(1))
        else:
            return None

    @sportsref.decorators.memoize
    def sos(self, year):
        """Returns the SOS (Strength of Schedule) for a team in a year, based
        on SRS.

        :year: The year for the season in question.
        :returns: A float of SOS.
        """
        try:
            sos_text = self._year_info_pq(year, "SOS").text()
        except ValueError:
            return None
        m = re.search(r"SOS\s*:\s*(\S+)", sos_text)
        if m:
            return float(m.group(1))
        else:
            return None

    @sportsref.decorators.memoize
    def off_coordinator(self, year):
        """Returns the coach ID for the team's OC in a given year.

        :year: An int representing the year.
        :returns: A string containing the coach ID of the OC.
        """
        try:
            oc_anchor = self._year_info_pq(year, "Offensive Coordinator")("a")
            if oc_anchor:
                return oc_anchor.attr["href"]
        except ValueError:
            return None

    @sportsref.decorators.memoize
    def def_coordinator(self, year):
        """Returns the coach ID for the team's DC in a given year.

        :year: An int representing the year.
        :returns: A string containing the coach ID of the DC.
        """
        try:
            dc_anchor = self._year_info_pq(year, "Defensive Coordinator")("a")
            if dc_anchor:
                return dc_anchor.attr["href"]
        except ValueError:
            return None

    @sportsref.decorators.memoize
    def stadium(self, year):
        """Returns the ID for the stadium in which the team played in a given
        year.

        :year: The year in question.
        :returns: A string representing the stadium ID.
        """
        anchor = self._year_info_pq(year, "Stadium")("a")
        return sportsref.utils.rel_url_to_id(anchor.attr["href"])

    @sportsref.decorators.memoize
    def off_scheme(self, year):
        """Returns the name of the offensive scheme the team ran in the given
        year.

        :year: Int representing the season year.
        :returns: A string representing the offensive scheme.
        """
        scheme_text = self._year_info_pq(year, "Offensive Scheme").text()
        m = re.search(r"Offensive Scheme[:\s]*(.+)\s*", scheme_text, re.I)
        if m:
            return m.group(1)
        else:
            return None

    @sportsref.decorators.memoize
    def def_alignment(self, year):
        """Returns the name of the defensive alignment the team ran in the
        given year.

        :year: Int representing the season year.
        :returns: A string representing the defensive alignment.
        """
        scheme_text = self._year_info_pq(year, "Defensive Alignment").text()
        m = re.search(r"Defensive Alignment[:\s]*(.+)\s*", scheme_text, re.I)
        if m:
            return m.group(1)
        else:
            return None

    @sportsref.decorators.memoize
    def team_stats(self, year):
        """Returns a Series (dict-like) of team stats from the team-season
        page.

        :year: Int representing the season.
        :returns: A Series of team stats.
        """
        doc = self.get_year_doc(year)
        table = doc("table#team_stats")
        df = sportsref.utils.parse_table(table)
        if df.empty:
            return pd.Series()
        return df.loc[df.player_id == "Team Stats"].iloc[0]

    @sportsref.decorators.memoize
    def opp_stats(self, year):
        """Returns a Series (dict-like) of the team's opponent's stats from the
        team-season page.

        :year: Int representing the season.
        :returns: A Series of team stats.
        """
        doc = self.get_year_doc(year)
        table = doc("table#team_stats")
        df = sportsref.utils.parse_table(table)
        return df.loc[df.player_id == "Opp. Stats"].iloc[0]

    @sportsref.decorators.memoize
    def passing(self, year):
        doc = self.get_year_doc(year)
        table = doc("table#passing")
        df = sportsref.utils.parse_table(table)
        return df

    @sportsref.decorators.memoize
    def rushing_and_receiving(self, year):
        doc = self.get_year_doc(year)
        table = doc("#rushing_and_receiving")
        df = sportsref.utils.parse_table(table)
        return df

    @sportsref.decorators.memoize
    def off_splits(self, year):
        """Returns a DataFrame of offensive team splits for a season.

        :year: int representing the season.
        :returns: Pandas DataFrame of split data.
        """
        doc = self.get_year_doc("{}_splits".format(year))
        tables = doc("table.stats_table")
        dfs = [sportsref.utils.parse_table(table) for table in list(tables.items())]
        dfs = [
            df.assign(split=df.columns[0]).rename(
                columns={df.columns[0]: "split_value"}
            )
            for df in dfs
        ]
        if not dfs:
            return pd.DataFrame()
        return pd.concat(dfs).reset_index(drop=True)

    @sportsref.decorators.memoize
    def def_splits(self, year):
        """Returns a DataFrame of defensive team splits (i.e. opponent splits)
        for a season.

        :year: int representing the season.
        :returns: Pandas DataFrame of split data.
        """
        doc = self.get_year_doc("{}_opp_splits".format(year))
        tables = doc("table.stats_table")
        dfs = [sportsref.utils.parse_table(table) for table in list(tables.items())]
        dfs = [
            df.assign(split=df.columns[0]).rename(
                columns={df.columns[0]: "split_value"}
            )
            for df in dfs
        ]
        if not dfs:
            return pd.DataFrame()
        return pd.concat(dfs).reset_index(drop=True)



import numpy as np
from pyquery import PyQuery as pq

import sportsref


class Team(object, metaclass=sportsref.decorators.Cached):
    def __init__(self, team_id):
        self.team_id = team_id.upper()

    def __eq__(self, other):
        return self.team_id == other.team_id

    def __hash__(self):
        return hash(self.team_id)

    @sportsref.decorators.memoize
    def team_year_url(self, yr_str):
        return f"{sportsref.nba.BASE_URL}/teams/{self.team_id}/{yr_str}.htm"

    @sportsref.decorators.memoize
    def get_main_doc(self):
        team_url = f"{sportsref.nba.BASE_URL}/teams/{self.team_id}"
        main_doc = pq(sportsref.utils.get_html(team_url))
        return main_doc

    @sportsref.decorators.memoize
    def get_year_doc(self, yr_str):
        return pq(sportsref.utils.get_html(self.team_year_url(yr_str)))

    @sportsref.decorators.memoize
    def name(self):
        """Returns the real name of the franchise given the team ID.

        Examples:
        'BOS' -> 'Boston Celtics'
        'NJN' -> 'Brooklyn Nets'

        :returns: A string corresponding to the team's full name.
        """
        doc = self.get_main_doc()
        name = doc('div#info h1[itemprop="name"]').text()
        return name

    @sportsref.decorators.memoize
    def roster(self, year):
        """Returns the roster table for the given year.

        :year: The year for which we want the roster; defaults to current year.
        :returns: A DataFrame containing roster information for that year.
        """
        doc = self.get_year_doc(year)
        table = doc("table#roster")
        df = sportsref.utils.parse_table(table)
        df["years_experience"] = (
            df["years_experience"].replace("R", 0).replace("", np.nan).astype(float)
        )
        return df

    # TODO: kind_rpb
    @sportsref.decorators.memoize
    def schedule(self, year):
        """Gets schedule information for a team-season.

        :year: The year for which we want the schedule.
        :returns: DataFrame of schedule information.
        """
        doc = self.get_year_doc(f"{year}_games")
        table = doc("table#games")
        df = sportsref.utils.parse_table(table)
        return df



import re

import numpy as np
import pandas as pd

import sportsref

HM_LINEUP_COLS = ["hm_player{}".format(i) for i in range(1, 6)]
AW_LINEUP_COLS = ["aw_player{}".format(i) for i in range(1, 6)]
ALL_LINEUP_COLS = AW_LINEUP_COLS + HM_LINEUP_COLS

PLAYER_RE = r"\w{0,7}\d{2}"

# parsing field goal attempts
shot_re = (
    rf"(?P<shooter>{PLAYER_RE}) "
    r"(?P<is_fgm>makes|misses) "
    r"(?P<is_three>2|3)\-pt "
    r"(?P<shot_type>jump shot|hook shot|layup|dunk) "
    r"(?:from (?P<shot_dist>\d+) ft|at rim)"
)
assist_re = rf" \(assist by (?P<assister>{PLAYER_RE})\)"
block_re = rf" \(block by (?P<blocker>{PLAYER_RE})\)"
SHOT_RE = re.compile(rf"{shot_re}(?:{assist_re}|{block_re})?", flags=re.I)

# parsing jump balls
jump_re = (
    rf"Jump ball: (?P<away_jumper>{PLAYER_RE}) vs\. (?P<home_jumper>{PLAYER_RE})"
    rf"(?: \((?P<gains_poss>{PLAYER_RE}) gains possession\))?"
)
JUMP_RE = re.compile(jump_re, flags=re.I)

# parsing rebounds
reb_re = rf"(?P<is_oreb>Offensive|Defensive) rebound by (?P<rebounder>{PLAYER_RE}|Team)"
REB_RE = re.compile(reb_re, flags=re.I)

# parsing free throws
ft_re = (
    rf"(?P<ft_shooter>{PLAYER_RE}) (?P<is_ftm>makes|misses) "
    r"(?P<is_tech_fta>technical )?(?P<is_flag_fta>flagrant )?"
    r"(?P<is_clearpath_fta>clear path )?free throw"
    r"(?: (?P<fta_num>\d+) of (?P<tot_fta>\d+))?"
)
FT_RE = re.compile(ft_re, flags=re.I)

# parsing substitutions
sub_re = rf"(?P<sub_in>{PLAYER_RE}) enters the game for (?P<sub_out>{PLAYER_RE})"
SUB_RE = re.compile(sub_re, flags=re.I)

# parsing turnovers
to_reasons = rf"(?P<to_type>[^;]+)(?:; steal by (?P<stealer>{PLAYER_RE}))?"
to_re = rf"Turnover by (?P<to_by>{PLAYER_RE}|Team) \((?:{to_reasons})\)"
TO_RE = re.compile(to_re, flags=re.I)

# parsing shooting fouls
shot_foul_re = (
    r"Shooting(?P<is_block_foul> block)? foul "
    rf"by (?P<fouler>{PLAYER_RE})"
    rf"(?: \(drawn by (?P<drew_foul>{PLAYER_RE})\))?"
)
SHOT_FOUL_RE = re.compile(shot_foul_re, flags=re.I)

# parsing offensive fouls
off_foul_re = (
    r"Offensive(?P<is_charge> charge)? foul "
    rf"by (?P<to_by>{PLAYER_RE})"
    rf"(?: \(drawn by (?P<drew_foul>{PLAYER_RE})\))?"
)
OFF_FOUL_RE = re.compile(off_foul_re, flags=re.I)

# parsing personal fouls
foul_re = (
    r"Personal (?P<is_take_foul>take )?(?P<is_block_foul>block )?"
    rf"foul by (?P<fouler>{PLAYER_RE})(?: \(drawn by "
    rf"(?P<drew_foul>{PLAYER_RE})\))?"
)
FOUL_RE = re.compile(foul_re, flags=re.I)

# parsing loose ball fouls
loose_ball_re = (
    rf"Loose ball foul by (?P<fouler>{PLAYER_RE})"
    rf"(?: \(drawn by (?P<drew_foul>{PLAYER_RE})\))?"
)
LOOSE_BALL_RE = re.compile(loose_ball_re, flags=re.I)

# parsing away from play fouls
away_from_ball_re = (
    rf"Away from play foul by (?P<fouler>{PLAYER_RE})"
    rf"(?: \(drawn by (?P<drew_foul>{PLAYER_RE})\))?"
)
AWAY_FROM_BALL_RE = re.compile(away_from_ball_re, flags=re.I)

# parsing inbound fouls
inbound_re = (
    rf"Inbound foul by (?P<fouler>{PLAYER_RE})"
    rf"(?: \(drawn by (?P<drew_foul>{PLAYER_RE})\))?"
)
INBOUND_RE = re.compile(inbound_re, flags=re.I)

# parsing flagrant fouls
flagrant_re = (
    rf"Flagrant foul type (?P<flag_type>1|2) by (?P<fouler>{PLAYER_RE})"
    rf"(?: \(drawn by (?P<drew_foul>{PLAYER_RE})\))?"
)
FLAGRANT_RE = re.compile(flagrant_re, flags=re.I)

# parsing clear path fouls
clear_path_re = (
    rf"Clear path foul by (?P<fouler>{PLAYER_RE})"
    rf"(?: \(drawn by (?P<drew_foul>{PLAYER_RE})\))?"
)
CLEAR_PATH_RE = re.compile(clear_path_re, flags=re.I)

# parsing timeouts
timeout_re = r"(?P<timeout_team>.*?) (?:full )?timeout"
TIMEOUT_RE = re.compile(timeout_re, flags=re.I)

# parsing technical fouls
tech_re = (
    r"(?P<is_hanging>Hanging )?"
    r"(?P<is_taunting>Taunting )?"
    r"(?P<is_ill_def>Ill def )?"
    r"(?P<is_delay>Delay )?"
    r"(?P<is_unsport>Non unsport )?"
    r"tech(?:nical)? foul by "
    rf"(?P<tech_fouler>{PLAYER_RE}|Team)"
)
TECH_RE = re.compile(tech_re, flags=re.I)

# parsing ejections
eject_re = rf"(?P<ejectee>{PLAYER_RE}|Team) ejected from game"
EJECT_RE = re.compile(eject_re, flags=re.I)

# parsing defensive 3 seconds techs
def3_tech_re = (
    r"(?:Def 3 sec tech foul|Defensive three seconds)"
    rf" by (?P<tech_fouler>{PLAYER_RE})"
)
DEF3_TECH_RE = re.compile(def3_tech_re, flags=re.I)

# parsing violations
viol_re = rf"Violation by (?P<violator>{PLAYER_RE}|Team) \((?P<viol_type>.*)\)"
VIOL_RE = re.compile(viol_re, flags=re.I)


def sparse_lineup_cols(df):
    regex = "{}_in".format(PLAYER_RE)
    return [c for c in df.columns if re.match(regex, c)]


def parse_play(boxscore_id, details, is_home):
    """Parse play details from a play-by-play string describing a play.

    Assuming valid input, this function returns structured data in a dictionary
    describing the play. If the play detail string was invalid, this function
    returns None.

    :param boxscore_id: the boxscore ID of the play
    :param details: detail string for the play
    :param is_home: bool indicating whether the offense is at home
    :param returns: dictionary of play attributes or None if invalid
    :rtype: dictionary or None
    """
    # if input isn't a string, return None
    if not details or not isinstance(details, str):
        return None

    bs = sportsref.nba.BoxScore(boxscore_id)
    aw, hm = bs.away(), bs.home()
    season = sportsref.nba.Season(bs.season())
    hm_roster = set(bs.basic_stats().query("is_home == True").player_id.values)

    play = {}
    play["detail"] = details
    play["home"] = hm
    play["away"] = aw
    play["is_home_play"] = is_home

    match = re.match(SHOT_RE, details)
    if match:
        play["is_fga"] = True
        play.update(match.groupdict())
        play["shot_dist"] = play["shot_dist"] if play["shot_dist"] is not None else 0
        play["shot_dist"] = int(play["shot_dist"])
        play["is_fgm"] = play["is_fgm"] == "makes"
        play["is_three"] = play["is_three"] == "3"
        play["is_assist"] = pd.notnull(play.get("assister"))
        play["is_block"] = pd.notnull(play.get("blocker"))
        shooter_home = play["shooter"] in hm_roster
        play["off_team"] = hm if shooter_home else aw
        play["def_team"] = aw if shooter_home else hm
        return play

    match = re.match(JUMP_RE, details)
    if match:
        play["is_jump_ball"] = True
        play.update(match.groupdict())
        return play

    match = re.match(REB_RE, details)
    if match:
        play["is_reb"] = True
        play.update(match.groupdict())
        play["is_oreb"] = play["is_oreb"].lower() == "offensive"
        play["is_dreb"] = not play["is_oreb"]
        play["is_team_rebound"] = play["rebounder"] == "Team"
        if play["is_team_rebound"]:
            play["reb_team"], other = (hm, aw) if is_home else (aw, hm)
        else:
            reb_home = play["rebounder"] in hm_roster
            play["reb_team"], other = (hm, aw) if reb_home else (aw, hm)
        play["off_team"] = play["reb_team"] if play["is_oreb"] else other
        play["def_team"] = play["reb_team"] if play["is_dreb"] else other
        return play

    match = re.match(FT_RE, details)
    if match:
        play["is_fta"] = True
        play.update(match.groupdict())
        play["is_ftm"] = play["is_ftm"] == "makes"
        play["is_tech_fta"] = bool(play["is_tech_fta"])
        play["is_flag_fta"] = bool(play["is_flag_fta"])
        play["is_clearpath_fta"] = bool(play["is_clearpath_fta"])
        play["is_pf_fta"] = not play["is_tech_fta"]
        if play["tot_fta"]:
            play["tot_fta"] = int(play["tot_fta"])
        if play["fta_num"]:
            play["fta_num"] = int(play["fta_num"])
        ft_home = play["ft_shooter"] in hm_roster
        play["fta_team"] = hm if ft_home else aw
        if not play["is_tech_fta"]:
            play["off_team"] = hm if ft_home else aw
            play["def_team"] = aw if ft_home else hm
        return play

    match = re.match(SUB_RE, details)
    if match:
        play["is_sub"] = True
        play.update(match.groupdict())
        sub_home = play["sub_in"] in hm_roster or play["sub_out"] in hm_roster
        play["sub_team"] = hm if sub_home else aw
        return play

    match = re.match(TO_RE, details)
    if match:
        play["is_to"] = True
        play.update(match.groupdict())
        play["to_type"] = play["to_type"].lower()
        if play["to_type"] == "offensive foul":
            return None
        play["is_steal"] = pd.notnull(play["stealer"])
        play["is_travel"] = play["to_type"] == "traveling"
        play["is_shot_clock_viol"] = play["to_type"] == "shot clock"
        play["is_oob"] = play["to_type"] == "step out of bounds"
        play["is_three_sec_viol"] = play["to_type"] == "3 sec"
        play["is_backcourt_viol"] = play["to_type"] == "back court"
        play["is_off_goaltend"] = play["to_type"] == "offensive goaltending"
        play["is_double_dribble"] = play["to_type"] == "dbl dribble"
        play["is_discont_dribble"] = play["to_type"] == "discontinued dribble"
        play["is_carry"] = play["to_type"] == "palming"
        if play["to_by"] == "Team":
            play["off_team"] = hm if is_home else aw
            play["def_team"] = aw if is_home else hm
        else:
            to_home = play["to_by"] in hm_roster
            play["off_team"] = hm if to_home else aw
            play["def_team"] = aw if to_home else hm
        return play

    match = re.match(SHOT_FOUL_RE, details)
    if match:
        play["is_pf"] = True
        play["is_shot_foul"] = True
        play.update(match.groupdict())
        play["is_block_foul"] = bool(play["is_block_foul"])
        foul_on_home = play["fouler"] in hm_roster
        play["off_team"] = aw if foul_on_home else hm
        play["def_team"] = hm if foul_on_home else aw
        play["foul_team"] = play["def_team"]
        return play

    match = re.match(OFF_FOUL_RE, details)
    if match:
        play["is_pf"] = True
        play["is_off_foul"] = True
        play["is_to"] = True
        play["to_type"] = "offensive foul"
        play.update(match.groupdict())
        play["is_charge"] = bool(play["is_charge"])
        play["fouler"] = play["to_by"]
        foul_on_home = play["fouler"] in hm_roster
        play["off_team"] = hm if foul_on_home else aw
        play["def_team"] = aw if foul_on_home else hm
        play["foul_team"] = play["off_team"]
        return play

    match = re.match(FOUL_RE, details)
    if match:
        play["is_pf"] = True
        play.update(match.groupdict())
        play["is_take_foul"] = bool(play["is_take_foul"])
        play["is_block_foul"] = bool(play["is_block_foul"])
        foul_on_home = play["fouler"] in hm_roster
        play["off_team"] = aw if foul_on_home else hm
        play["def_team"] = hm if foul_on_home else aw
        play["foul_team"] = play["def_team"]
        return play

    # TODO: parsing double personal fouls
    # double_foul_re = (r'Double personal foul by (?P<fouler1>{0}) and '
    #                   r'(?P<fouler2>{0})').format(PLAYER_RE)
    # m = re.match(double_Foul_re, details)
    # if m:
    #     p['is_pf'] = True
    #     p.update(m.groupdict())
    #     p['off_team'] =

    match = re.match(LOOSE_BALL_RE, details)
    if match:
        play["is_pf"] = True
        play["is_loose_ball_foul"] = True
        play.update(match.groupdict())
        foul_home = play["fouler"] in hm_roster
        play["foul_team"] = hm if foul_home else aw
        return play

    # parsing punching fouls
    # TODO

    match = re.match(AWAY_FROM_BALL_RE, details)
    if match:
        play["is_pf"] = True
        play["is_away_from_play_foul"] = True
        play.update(match.groupdict())
        foul_on_home = play["fouler"] in hm_roster
        # TODO: figure out who had the ball based on previous play
        play["foul_team"] = hm if foul_on_home else aw
        return play

    match = re.match(INBOUND_RE, details)
    if match:
        play["is_pf"] = True
        play["is_inbound_foul"] = True
        play.update(match.groupdict())
        foul_on_home = play["fouler"] in hm_roster
        play["off_team"] = aw if foul_on_home else hm
        play["def_team"] = hm if foul_on_home else aw
        play["foul_team"] = play["def_team"]
        return play

    match = re.match(FLAGRANT_RE, details)
    if match:
        play["is_pf"] = True
        play["is_flagrant"] = True
        play.update(match.groupdict())
        foul_on_home = play["fouler"] in hm_roster
        play["foul_team"] = hm if foul_on_home else aw
        return play

    match = re.match(CLEAR_PATH_RE, details)
    if match:
        play["is_pf"] = True
        play["is_clear_path_foul"] = True
        play.update(match.groupdict())
        foul_on_home = play["fouler"] in hm_roster
        play["off_team"] = aw if foul_on_home else hm
        play["def_team"] = hm if foul_on_home else aw
        play["foul_team"] = play["def_team"]
        return play

    match = re.match(TIMEOUT_RE, details)
    if match:
        play["is_timeout"] = True
        play.update(match.groupdict())
        isOfficialTO = play["timeout_team"].lower() == "official"
        name_to_id = season.team_names_to_ids()
        play["timeout_team"] = (
            "Official"
            if isOfficialTO
            else name_to_id.get(hm, name_to_id.get(aw, play["timeout_team"]))
        )
        return play

    match = re.match(TECH_RE, details)
    if match:
        play["is_tech_foul"] = True
        play.update(match.groupdict())
        play["is_hanging"] = bool(play["is_hanging"])
        play["is_taunting"] = bool(play["is_taunting"])
        play["is_ill_def"] = bool(play["is_ill_def"])
        play["is_delay"] = bool(play["is_delay"])
        play["is_unsport"] = bool(play["is_unsport"])
        foul_on_home = play["tech_fouler"] in hm_roster
        play["foul_team"] = hm if foul_on_home else aw
        return play

    match = re.match(EJECT_RE, details)
    if match:
        play["is_ejection"] = True
        play.update(match.groupdict())
        if play["ejectee"] == "Team":
            play["ejectee_team"] = hm if is_home else aw
        else:
            eject_home = play["ejectee"] in hm_roster
            play["ejectee_team"] = hm if eject_home else aw
        return play

    match = re.match(DEF3_TECH_RE, details)
    if match:
        play["is_tech_foul"] = True
        play["is_def_three_secs"] = True
        play.update(match.groupdict())
        foul_on_home = play["tech_fouler"] in hm_roster
        play["off_team"] = aw if foul_on_home else hm
        play["def_team"] = hm if foul_on_home else aw
        play["foul_team"] = play["def_team"]
        return play

    match = re.match(VIOL_RE, details)
    if match:
        play["is_viol"] = True
        play.update(match.groupdict())
        if play["viol_type"] == "kicked_ball":
            play["is_to"] = True
            play["to_by"] = play["violator"]
        if play["violator"] == "Team":
            play["viol_team"] = hm if is_home else aw
        else:
            viol_home = play["violator"] in hm_roster
            play["viol_team"] = hm if viol_home else aw
        return play

    play["is_error"] = True
    return play


def clean_features(df):
    """Fixes up columns of the passed DataFrame, such as casting T/F columns to
    boolean and filling in NaNs for team and opp.

    :param df: DataFrame of play-by-play data.
    :returns: Dataframe with cleaned columns.
    """
    df = pd.DataFrame(df)

    bool_vals = set([True, False, None, np.nan])
    sparse_cols = sparse_lineup_cols(df)
    for col in df:

        # make indicator columns boolean type (and fill in NaNs)
        if set(df[col].unique()[:5]) <= bool_vals:
            df[col] = df[col] == True  # noqa

        # fill NaN's in sparse lineup columns to 0
        elif col in sparse_cols:
            df[col] = df[col].fillna(0)

    # fix free throw columns on technicals
    df.loc[df.is_tech_fta, ["fta_num", "tot_fta"]] = 1

    # fill in NaN's/fix off_team and def_team columns
    df.off_team.fillna(method="bfill", inplace=True)
    df.def_team.fillna(method="bfill", inplace=True)
    df.off_team.fillna(method="ffill", inplace=True)
    df.def_team.fillna(method="ffill", inplace=True)

    return df


def clean_multigame_features(df):
    """TODO: Docstring for clean_multigame_features.

    :df: TODO
    :returns: TODO
    """
    df = pd.DataFrame(df)
    if df.index.value_counts().max() > 1:
        df.reset_index(drop=True, inplace=True)

    df = clean_features(df)

    # if it's many games in one DataFrame, make poss_id and play_id unique
    for col in ("play_id", "poss_id"):
        diffs = df[col].diff().fillna(0)
        if (diffs < 0).any():
            new_col = np.cumsum(diffs.astype(bool))  # noqa
            df.eval("{} = @new_col".format(col), inplace=True)

    return df


def get_period_starters(df):
    """TODO"""

    def players_from_play(play):
        """Figures out what players are in the game based on the players
        mentioned in a play. Returns away and home players as two sets.

        :param play: A dictionary representing a parsed play.
        :returns: (aw_players, hm_players)
        :rtype: tuple of lists
        """
        # if it's a tech FT from between periods, don't count this play
        if play["clock_str"] == "12:00.0" and (
            play.get("is_tech_foul") or play.get("is_tech_fta")
        ):
            return [], []

        stats = sportsref.nba.BoxScore(play["boxscore_id"]).basic_stats()
        home_grouped = stats.groupby("is_home")
        hm_roster = set(home_grouped.player_id.get_group(True).values)
        aw_roster = set(home_grouped.player_id.get_group(False).values)
        player_keys = [
            "assister",
            "away_jumper",
            "blocker",
            "drew_foul",
            "fouler",
            "ft_shooter",
            "gains_poss",
            "home_jumper",
            "rebounder",
            "shooter",
            "stealer",
            "sub_in",
            "sub_out",
            "to_by",
        ]
        players = [p for p in play[player_keys] if pd.notnull(p)]

        aw_players = [p for p in players if p in aw_roster]
        hm_players = [p for p in players if p in hm_roster]
        return aw_players, hm_players

    # create a mapping { quarter => (away_starters, home_starters) }
    n_periods = df.quarter.nunique()
    period_starters = [(set(), set()) for _ in range(n_periods)]

    # fill out this mapping quarter by quarter
    for qtr, qtr_grp in df.groupby(df.quarter):
        aw_starters, hm_starters = period_starters[qtr - 1]
        exclude = set()
        # loop through sets of plays that happen at the "same time"
        for label, time_grp in qtr_grp.groupby(qtr_grp.secs_elapsed):
            # first, if they sub in and weren't already starters, exclude them
            sub_ins = set(time_grp.sub_in.dropna().values)
            exclude.update(sub_ins - aw_starters - hm_starters)
            # second, figure out new starters from each play at this time
            for i, row in time_grp.iterrows():
                aw_players, hm_players = players_from_play(row)
                # update overall sets for the quarter
                aw_starters.update(aw_players)
                hm_starters.update(hm_players)
            # remove excluded (subbed-in) players
            hm_starters -= exclude
            aw_starters -= exclude
            # check whether we have found all starters
            if len(hm_starters) > 5 or len(aw_starters) > 5:
                import ipdb

                ipdb.set_trace()
            if len(hm_starters) >= 5 and len(aw_starters) >= 5:
                break

        if len(hm_starters) != 5 or len(aw_starters) != 5:
            print(
                "WARNING: wrong number of starters for a team in Q{} of {}".format(
                    qtr, df.boxscore_id.iloc[0]
                )
            )

    return period_starters


def get_sparse_lineups(df):
    """TODO: Docstring for get_sparse_lineups.

    :param df: TODO
    :returns: TODO
    """

    # get the lineup data using get_dense_lineups if necessary
    if set(ALL_LINEUP_COLS) - set(df.columns):
        lineup_df = get_dense_lineups(df)
    else:
        lineup_df = df[ALL_LINEUP_COLS]

    # create the sparse representation
    hm_lineups = lineup_df[HM_LINEUP_COLS].values
    aw_lineups = lineup_df[AW_LINEUP_COLS].values
    # +1 for home, -1 for away
    hm_df = pd.DataFrame(
        [
            {"{}_in".format(player_id): 1 for player_id in lineup}
            for lineup in hm_lineups
        ],
        dtype=int,
    )
    aw_df = pd.DataFrame(
        [
            {"{}_in".format(player_id): -1 for player_id in lineup}
            for lineup in aw_lineups
        ],
        dtype=int,
    )
    sparse_df = pd.concat((hm_df, aw_df), axis=1).fillna(0)
    return sparse_df


def get_dense_lineups(df):
    """Returns a new DataFrame based on the one it is passed. Specifically, it
    adds five columns for each team (ten total), where each column has the ID
    of a player on the court during the play.

    This information is figured out sequentially from the game's substitution
    data in the passed DataFrame, so the DataFrame passed as an argument must
    be from a specific BoxScore (rather than a DataFrame of non-consecutive
    plays). That is, the DataFrame must be of the form returned by
    :func:`nba.BoxScore.pbp <nba.BoxScore.pbp>`.

    .. note:: Note that the lineups reflect the teams in the game when the play
        happened, not after the play. For example, if a play is a substitution,
        the lineups for that play will be the lineups before the substituion
        occurs.

    :param df: A DataFrame of a game's play-by-play data.
    :returns: A DataFrame with additional lineup columns.

    """
    # TODO: add this precondition to documentation
    assert df["boxscore_id"].nunique() == 1

    def lineup_dict(aw_lineup, hm_lineup):
        """Returns a dictionary of lineups to be converted to columns.
        Specifically, the columns are 'aw_player1' through 'aw_player5' and
        'hm_player1' through 'hm_player5'.

        :param aw_lineup: The away team's current lineup.
        :param hm_lineup: The home team's current lineup.
        :returns: A dictionary of lineups.
        """
        return {
            "{}_player{}".format(tm, i + 1): player
            for tm, lineup in zip(["aw", "hm"], [aw_lineup, hm_lineup])
            for i, player in enumerate(lineup)
        }

    def handle_sub(row, aw_lineup, hm_lineup):
        """Modifies the aw_lineup and hm_lineup lists based on the substitution
        that takes place in the given row."""
        assert row["is_sub"]
        sub_lineup = hm_lineup if row["sub_team"] == row["home"] else aw_lineup
        try:
            # make the sub
            idx = sub_lineup.index(row["sub_out"])
            sub_lineup[idx] = row["sub_in"]
        except ValueError:
            # if the sub was double-entered and it's already been executed...
            if row["sub_in"] in sub_lineup and row["sub_out"] not in sub_lineup:
                return aw_lineup, hm_lineup
            # otherwise, let's print and pretend this never happened
            print(
                "ERROR IN SUB IN {}, Q{}, {}: {}".format(
                    row["boxscore_id"], row["quarter"], row["clock_str"], row["detail"]
                )
            )
            raise
        return aw_lineup, hm_lineup

    per_starters = get_period_starters(df)
    cur_qtr = 0
    aw_lineup, hm_lineup = [], []
    df = df.reset_index(drop=True)
    lineups = [{} for _ in range(df.shape[0])]

    # loop through select plays to determine lineups
    sub_or_per_start = df.is_sub | df.quarter.diff().astype(bool)
    for i, row in df.loc[sub_or_per_start].iterrows():
        if row["quarter"] > cur_qtr:
            # first row in a quarter
            assert row["quarter"] == cur_qtr + 1
            # first, finish up the last quarter's lineups
            if cur_qtr > 0 and not df.loc[i - 1, "is_sub"]:
                lineups[i - 1] = lineup_dict(aw_lineup, hm_lineup)
            # then, move on to the quarter, and enter the starting lineups
            cur_qtr += 1
            aw_lineup, hm_lineup = list(map(list, per_starters[cur_qtr - 1]))
            lineups[i] = lineup_dict(aw_lineup, hm_lineup)
            # if the first play in the quarter is a sub, handle that
            if row["is_sub"]:
                aw_lineup, hm_lineup = handle_sub(row, aw_lineup, hm_lineup)
        else:
            # during the quarter
            # update lineups first then change lineups based on subs
            lineups[i] = lineup_dict(aw_lineup, hm_lineup)
            if row["is_sub"]:
                aw_lineup, hm_lineup = handle_sub(row, aw_lineup, hm_lineup)

    # create and clean DataFrame
    lineup_df = pd.DataFrame(lineups)
    if lineup_df.iloc[-1].isnull().all():
        lineup_df.iloc[-1] = lineup_dict(aw_lineup, hm_lineup)
    lineup_df = lineup_df.groupby(df.quarter).fillna(method="bfill")

    # fill in NaN's based on minutes played
    bool_mat = lineup_df.isnull()
    mask = bool_mat.any(axis=1)
    if mask.any():
        bs = sportsref.nba.BoxScore(df.boxscore_id[0])
        # first, get the true minutes played from the box score
        stats = sportsref.nba.BoxScore(df.boxscore_id.iloc[0]).basic_stats()
        true_mp = (
            pd.Series(
                stats.query("mp > 0")[["player_id", "mp"]]
                .set_index("player_id")
                .to_dict()["mp"]
            )
            * 60
        )
        # next, calculate minutes played based on the lineup data
        calc_mp = pd.Series(
            {
                p: (
                    df.secs_elapsed.diff() * [p in row for row in lineup_df.values]
                ).sum()
                for p in stats.query("mp > 0").player_id.values
            }
        )
        # finally, figure which players are missing minutes
        diff = true_mp - calc_mp
        players_missing = diff.loc[diff.abs() >= 150]
        hm_roster = bs.basic_stats().query("is_home == True").player_id.values
        missing_df = pd.DataFrame(
            {
                "secs": players_missing.values,
                "is_home": players_missing.index.isin(hm_roster),
            },
            index=players_missing.index,
        )

        if missing_df.empty:
            # TODO: log this as a warning (or error?)
            print(
                "There are NaNs in the lineup data, but no players were "
                "found to be missing significant minutes"
            )
        else:
            # import ipdb
            # ipdb.set_trace()
            for is_home, group in missing_df.groupby("is_home"):
                player_id = group.index.item()
                tm_cols = (
                    sportsref.nba.pbp.HM_LINEUP_COLS
                    if is_home
                    else sportsref.nba.pbp.AW_LINEUP_COLS
                )
                row_mask = lineup_df[tm_cols].isnull().any(axis=1)
                lineup_df.loc[row_mask, tm_cols] = (
                    lineup_df.loc[row_mask, tm_cols].fillna(player_id).values
                )

    return lineup_df



import datetime
import re

from pyquery import PyQuery as pq

import sportsref

__all__ = ["Player"]


class Player(object, metaclass=sportsref.decorators.Cached):

    """Each instance of this class represents an NBA player, uniquely
    identified by a player ID. The instance methods give various data available
    from the player's Basketball Reference player page."""

    def __init__(self, player_id):
        self.player_id = player_id
        self.url_base = f"{sportsref.nba.BASE_URL}/players/{player_id[0]}/{player_id}"
        self.main_url = self.url_base + ".htm"

    def __eq__(self, other):
        return self.player_id == other.player_id

    def __hash__(self):
        return hash(self.player_id)

    def __repr__(self):
        return f"Player({self.player_id})"

    def __str__(self):
        return self.name()

    @sportsref.decorators.memoize
    def get_main_doc(self):
        return pq(sportsref.utils.get_html(self.main_url))

    @sportsref.decorators.memoize
    def get_sub_doc(self, rel_url):
        url = f"{self.url_base}/{rel_url}"
        return pq(sportsref.utils.get_html(url))

    @sportsref.decorators.memoize
    def name(self):
        """Returns the name of the player as a string."""
        doc = self.get_main_doc()
        return doc('h1[itemprop="name"]').text()

    @sportsref.decorators.memoize
    def age(self, year, month=2, day=1):
        """Returns the age of the player on a given date.

        :year: int representing the year.
        :month: int representing the month (1-12).
        :day: int representing the day within the month (1-31).
        :returns: Age in years as a float.
        """
        doc = self.get_main_doc()
        date_string = doc('span[itemprop="birthDate"]').attr("data-birth")
        regex = r"(\d{4})\-(\d{2})\-(\d{2})"
        date_args = list(map(int, re.match(regex, date_string).groups()))
        birth_date = datetime.date(*date_args)
        age_date = datetime.date(year=year, month=month, day=day)
        delta = age_date - birth_date
        age = delta.days / 365.0
        return age

    @sportsref.decorators.memoize
    def position(self):
        """TODO: Docstring for position.
        :returns: TODO
        """
        raise Exception("not yet implemented - nba.Player.position")

    @sportsref.decorators.memoize
    def height(self):
        """Returns the player's height (in inches).
        :returns: An int representing a player's height in inches.
        """
        doc = self.get_main_doc()
        raw = doc('span[itemprop="height"]').text()
        try:
            feet, inches = list(map(int, raw.split("-")))
            return feet * 12 + inches
        except ValueError:
            return None

    @sportsref.decorators.memoize
    def weight(self):
        """Returns the player's weight (in pounds).
        :returns: An int representing a player's weight in pounds.
        """
        doc = self.get_main_doc()
        raw = doc('span[itemprop="weight"]').text()
        try:
            weight = re.match(r"(\d+)lb", raw).group(1)
            return int(weight)
        except ValueError:
            return None

    @sportsref.decorators.memoize
    def hand(self):
        """Returns the player's handedness.
        :returns: 'L' for left-handed, 'R' for right-handed.
        """
        doc = self.get_main_doc()
        hand = re.search(r"Shoots:\s*(L|R)", doc.text()).group(1)
        return hand

    @sportsref.decorators.memoize
    def draft_pick(self):
        """Returns when in the draft the player was picked.
        :returns: TODO
        """
        doc = self.get_main_doc()
        try:
            p_tags = doc("div#meta p")
            draft_p_tag = next(
                p for p in list(p_tags.items()) if p.text().lower().startswith("draft")
            )
            draft_pick = int(
                re.search(r"(\d+)\w{,3}\s+?overall", draft_p_tag.text()).group(1)
            )
            return draft_pick
        except Exception:
            return None

    @sportsref.decorators.memoize
    def draft_year(self):
        """Returns the year the player was selected (or undrafted).
        :returns: TODO
        """
        raise Exception("not yet implemented - nba.Player.draft_year")

    @sportsref.decorators.kind_rpb(include_type=True)
    def _get_stats_table(self, table_id, kind="R", summary=False):
        """Gets a stats table from the player page; helper function that does
        the work for per-game, per-100-poss, etc. stats.

        :table_id: the ID of the HTML table.
        :kind: specifies regular season, playoffs, or both. One of 'R', 'P',
            'B'. Defaults to 'R'.
        :returns: A DataFrame of stats.
        """
        doc = self.get_main_doc()
        table_id = f"table#{'playoffs_' if kind == 'P' else ''}{table_id}"
        table = doc(table_id)
        df = sportsref.utils.parse_table(table, flatten=(not summary), footer=summary)
        return df

    @sportsref.decorators.memoize
    def stats_per_game(self, kind="R", summary=False):
        """Returns a DataFrame of per-game box score stats."""
        return self._get_stats_table("per_game", kind=kind, summary=summary)

    @sportsref.decorators.memoize
    def stats_totals(self, kind="R", summary=False):
        """Returns a DataFrame of total box score statistics by season."""
        return self._get_stats_table("totals", kind=kind, summary=summary)

    @sportsref.decorators.memoize
    def stats_per36(self, kind="R", summary=False):
        """Returns a DataFrame of per-36-minutes stats."""
        return self._get_stats_table("per_minute", kind=kind, summary=summary)

    @sportsref.decorators.memoize
    def stats_per100(self, kind="R", summary=False):
        """Returns a DataFrame of per-100-possession stats."""
        return self._get_stats_table("per_poss", kind=kind, summary=summary)

    @sportsref.decorators.memoize
    def stats_advanced(self, kind="R", summary=False):
        """Returns a DataFrame of advanced stats."""
        return self._get_stats_table("advanced", kind=kind, summary=summary)

    @sportsref.decorators.memoize
    def stats_shooting(self, kind="R", summary=False):
        """Returns a DataFrame of shooting stats."""
        return self._get_stats_table("shooting", kind=kind, summary=summary)

    @sportsref.decorators.memoize
    def stats_adjusted_shooting(self, kind="R", summary=False):
        """Returns a DataFrame of adjusted shooting stats."""
        return self._get_stats_table("adj-shooting", kind=kind, summary=summary)

    @sportsref.decorators.memoize
    def stats_pbp(self, kind="R", summary=False):
        """Returns a DataFrame of play-by-play stats."""
        return self._get_stats_table("pbp", kind=kind, summary=summary)

    @sportsref.decorators.memoize
    @sportsref.decorators.kind_rpb(include_type=True)
    def gamelog_basic(self, year, kind="R"):
        """Returns a table of a player's basic game-by-game stats for a season.

        :param year: The year representing the desired season.
        :param kind: specifies regular season, playoffs, or both. One of 'R',
            'P', 'B'. Defaults to 'R'.
        :returns: A DataFrame of the player's standard boxscore stats from each
            game of the season.
        :rtype: pd.DataFrame
        """
        doc = self.get_sub_doc(f"gamelog/{year}")
        table = (
            doc("table#pgl_basic_playoffs") if kind == "P" else doc("table#pgl_basic")
        )
        df = sportsref.utils.parse_table(table)
        return df

    @sportsref.decorators.memoize
    @sportsref.decorators.kind_rpb(include_type=True)
    def gamelog_advanced(self, year, kind="R"):
        """Returns a table of a player's advanced game-by-game stats for a
        season.

        :param year: The year representing the desired season.
        :param kind: specifies regular season, playoffs, or both. One of 'R',
            'P', 'B'. Defaults to 'R'.
        :returns: A DataFrame of the player's advanced stats from each game of
            the season.
        :rtype: pd.DataFrame
        """
        doc = self.get_sub_doc(f"gamelog-advanced/{year}")
        table = (
            doc("table#pgl_advanced_playoffs")
            if kind == "P"
            else doc("table#pgl_advanced")
        )
        df = sportsref.utils.parse_table(table)
        return df



from . import GPF
from . import PSF

from .PSF import PlayerSeasonFinder
from .GPF import GamePlayFinder

# modules/variables to expose
__all__ = ["PlayerSeasonFinder", "GamePlayFinder"]

# Fill in PlayerSeasonFinder docstring

IOD = PSF.inputs_options_defaults()

paramStr = "\n".join(
    ':param {}: default="{}"'.format(name, ",".join(dct["value"]))
    for name, dct in sorted(IOD.items())
)
optsStr = "\n".join(
    "{}: {}".format(name, ",".join('"{}"'.format(opt) for opt in dct["options"]))
    if len(dct["options"]) <= 10
    else "{}: {}...{}".format(
        name,
        ",".join('"{}"'.format(opt) for opt in dct["options"][:10]),
        ",".join('"{}"'.format(opt) for opt in dct["options"][-2:]),
    )
    for name, dct in sorted(IOD.items())
)


PSF.PlayerSeasonFinder.__doc__ = """
Finds player-seasons that match criteria supplied by keyword arguments.

* Can use tm or team for team_id.
* Can use yr, year, yrs, or years for year_min, year_max.
* Can use [draft_]pos, [draft_]position, [draft_]positions for a shortcut for
[draft_]positions.

Options for inputs:
{}

{}
:returns: list of matching player-season tuples
:rtype: [(player ID, season year)]

""".format(
    paramStr, optsStr
)

# clean up namespace
del IOD, paramStr, optsStr


# Fill in GamePlayFinder docstring

IOD = GPF.inputs_options_defaults()

paramStr = "\n".join(
    ':param {}: default="{}"'.format(name, ",".join(dct["value"]))
    for name, dct in sorted(IOD.items())
)

optsStr = "\n".join(
    "{}: {}".format(name, ",".join('"{}"'.format(opt) for opt in dct["options"]))
    if len(dct["options"]) <= 10
    else "{}: {}...{}".format(
        name,
        ",".join('"{}"'.format(opt) for opt in dct["options"][:10]),
        ",".join('"{}"'.format(opt) for opt in dct["options"][-2:]),
    )
    for name, dct in sorted(IOD.items())
)

GPF.GamePlayFinder.__doc__ = """
Finds plays that match criteria supplied by keyword arguments.

* Can use tm or team instead of team_id.
* Can use yr, year, yrs, or years instead of year_min, year_max.
* For multi-valued options (like down or rush direction), separate values with
commas or use a list.
* For options that are yes/no/either or yes/no/any, -1 is either/any, 0 is no,
1 is yes.

Options for the inputs:
{}

{}
:returns: Pandas dataframe of plays
:rtype: pd.DataFrame
""".format(
    paramStr, optsStr
)

# clean up namespace
del IOD, paramStr, optsStr



import datetime
import functools
import re

import numpy as np
import pandas as pd
from pyquery import PyQuery as pq

import sportsref

__all__ = ["BoxScore"]


class BoxScore(object, metaclass=sportsref.decorators.Cached):
    def __init__(self, boxscore_id):
        self.boxscore_id = boxscore_id

    def __eq__(self, other):
        return self.boxscore_id == other.boxscore_id

    def __hash__(self):
        return hash(self.boxscore_id)

    def __repr__(self):
        return "BoxScore({})".format(self.boxscore_id)

    def __str__(self):
        return "{} Week {}: {} @ {}".format(
            self.season(), self.week(), self.away(), self.home()
        )

    def __reduce__(self):
        return BoxScore, (self.boxscore_id,)

    @sportsref.decorators.memoize
    def get_doc(self):
        url = sportsref.nfl.BASE_URL + "/boxscores/{}.htm".format(self.boxscore_id)
        doc = pq(sportsref.utils.get_html(url))
        return doc

    @sportsref.decorators.memoize
    def date(self):
        """Returns the date of the game. See Python datetime.date documentation
        for more.
        :returns: A datetime.date object with year, month, and day attributes.
        """
        match = re.match(r"(\d{4})(\d{2})(\d{2})", self.boxscore_id)
        year, month, day = list(map(int, match.groups()))
        return datetime.date(year=year, month=month, day=day)

    @sportsref.decorators.memoize
    def weekday(self):
        """Returns the day of the week on which the game occurred.
        :returns: String representation of the day of the week for the game.

        """
        days = [
            "Monday",
            "Tuesday",
            "Wednesday",
            "Thursday",
            "Friday",
            "Saturday",
            "Sunday",
        ]
        date = self.date()
        wd = date.weekday()
        return days[wd]

    @sportsref.decorators.memoize
    def home(self):
        """Returns home team ID.
        :returns: 3-character string representing home team's ID.
        """
        doc = self.get_doc()
        table = doc("table.linescore")
        relURL = table("tr").eq(2)("a").eq(2).attr["href"]
        home = sportsref.utils.rel_url_to_id(relURL)
        return home

    @sportsref.decorators.memoize
    def away(self):
        """Returns away team ID.
        :returns: 3-character string representing away team's ID.
        """
        doc = self.get_doc()
        table = doc("table.linescore")
        relURL = table("tr").eq(1)("a").eq(2).attr["href"]
        away = sportsref.utils.rel_url_to_id(relURL)
        return away

    @sportsref.decorators.memoize
    def home_score(self):
        """Returns score of the home team.
        :returns: int of the home score.
        """
        doc = self.get_doc()
        table = doc("table.linescore")
        home_score = table("tr").eq(2)("td")[-1].text_content()
        return int(home_score)

    @sportsref.decorators.memoize
    def away_score(self):
        """Returns score of the away team.
        :returns: int of the away score.
        """
        doc = self.get_doc()
        table = doc("table.linescore")
        away_score = table("tr").eq(1)("td")[-1].text_content()
        return int(away_score)

    @sportsref.decorators.memoize
    def winner(self):
        """Returns the team ID of the winning team. Returns NaN if a tie."""
        hmScore = self.home_score()
        awScore = self.away_score()
        if hmScore > awScore:
            return self.home()
        elif hmScore < awScore:
            return self.away()
        else:
            return None

    @sportsref.decorators.memoize
    def week(self):
        """Returns the week in which this game took place. 18 is WC round, 19
        is Div round, 20 is CC round, 21 is SB.
        :returns: Integer from 1 to 21.
        """
        doc = self.get_doc()
        raw = doc("div#div_other_scores h2 a").attr["href"]
        match = re.match(r"/years/{}/week_(\d+)\.htm".format(self.season()), raw)
        if match:
            return int(match.group(1))
        else:
            return 21  # super bowl is week 21

    @sportsref.decorators.memoize
    def season(self):
        """
        Returns the year ID of the season in which this game took place.
        Useful for week 17 January games.

        :returns: An int representing the year of the season.
        """
        date = self.date()
        return date.year - 1 if date.month <= 3 else date.year

    @sportsref.decorators.memoize
    def starters(self):
        """Returns a DataFrame where each row is an entry in the starters table
        from PFR.

        The columns are:
        * player_id - the PFR player ID for the player (note that this column
        is not necessarily all unique; that is, one player can be a starter in
        multiple positions, in theory).
        * playerName - the listed name of the player; this too is not
        necessarily unique.
        * position - the position at which the player started for their team.
        * team - the team for which the player started.
        * home - True if the player's team was at home, False if they were away
        * offense - True if the player is starting on an offensive position,
        False if defense.

        :returns: A pandas DataFrame. See the description for details.
        """
        doc = self.get_doc()
        a = doc("table#vis_starters")
        h = doc("table#home_starters")
        data = []
        for h, table in enumerate((a, h)):
            team = self.home() if h else self.away()
            for i, row in enumerate(table("tbody tr").items()):
                datum = {}
                datum["player_id"] = sportsref.utils.rel_url_to_id(
                    row("a")[0].attrib["href"]
                )
                datum["playerName"] = row("th").text()
                datum["position"] = row("td").text()
                datum["team"] = team
                datum["home"] = h == 1
                datum["offense"] = i <= 10
                data.append(datum)
        return pd.DataFrame(data)

    @sportsref.decorators.memoize
    def line(self):
        doc = self.get_doc()
        table = doc("table#game_info")
        giTable = sportsref.utils.parse_info_table(table)
        line_text = giTable.get("vegas_line", None)
        if line_text is None:
            return None
        m = re.match(r"(.+?) ([\-\.\d]+)$", line_text)
        if m:
            favorite, line = m.groups()
            line = float(line)
            # give in terms of the home team
            year = self.season()
            if favorite != sportsref.nfl.teams.team_names(year)[self.home()]:
                line = -line
        else:
            line = 0
        return line

    @sportsref.decorators.memoize
    def surface(self):
        """The playing surface on which the game was played.

        :returns: string representing the type of surface. Returns np.nan if
        not avaiable.
        """
        doc = self.get_doc()
        table = doc("table#game_info")
        giTable = sportsref.utils.parse_info_table(table)
        return giTable.get("surface", np.nan)

    @sportsref.decorators.memoize
    def over_under(self):
        """
        Returns the over/under for the game as a float, or np.nan if not
        available.
        """
        doc = self.get_doc()
        table = doc("table#game_info")
        giTable = sportsref.utils.parse_info_table(table)
        if "over_under" in giTable:
            ou = giTable["over_under"]
            return float(ou.split()[0])
        else:
            return None

    @sportsref.decorators.memoize
    def coin_toss(self):
        """Gets information relating to the opening coin toss.

        Keys are:
        * wonToss - contains the ID of the team that won the toss
        * deferred - bool whether the team that won the toss deferred it

        :returns: Dictionary of coin toss-related info.
        """
        doc = self.get_doc()
        table = doc("table#game_info")
        giTable = sportsref.utils.parse_info_table(table)
        if "Won Toss" in giTable:
            # TODO: finish coinToss function
            pass
        else:
            return None

    @sportsref.decorators.memoize
    def weather(self):
        """Returns a dictionary of weather-related info.

        Keys of the returned dict:
        * temp
        * windChill
        * relHumidity
        * windMPH

        :returns: Dict of weather data.
        """
        doc = self.get_doc()
        table = doc("table#game_info")
        giTable = sportsref.utils.parse_info_table(table)
        if "weather" in giTable:
            regex = (
                r"(?:(?P<temp>\-?\d+) degrees )?"
                r"(?:relative humidity (?P<relHumidity>\d+)%, )?"
                r"(?:wind (?P<windMPH>\d+) mph, )?"
                r"(?:wind chill (?P<windChill>\-?\d+))?"
            )
            m = re.match(regex, giTable["weather"])
            d = m.groupdict()

            # cast values to int
            for k in d:
                try:
                    d[k] = int(d[k])
                except TypeError:
                    pass

            # one-off fixes
            d["windChill"] = d["windChill"] if pd.notnull(d["windChill"]) else d["temp"]
            d["windMPH"] = d["windMPH"] if pd.notnull(d["windMPH"]) else 0
            return d
        else:
            # no weather found, because it's a dome
            # TODO: what's relative humidity in a dome?
            return {"temp": 70, "windChill": 70, "relHumidity": None, "windMPH": 0}

    @sportsref.decorators.memoize
    def pbp(self):
        """Returns a dataframe of the play-by-play data from the game.

        Order of function calls:
            1. parse_table on the play-by-play table
            2. expand_details
                - calls parse_play_details & _clean_features
            3. _add_team_columns
            4. various fixes to clean data
            5. _add_team_features

        :returns: pandas DataFrame of play-by-play. Similar to GPF.
        """
        doc = self.get_doc()
        table = doc("table#pbp")
        df = sportsref.utils.parse_table(table)
        # make the following features conveniently available on each row
        df["boxscore_id"] = self.boxscore_id
        df["home"] = self.home()
        df["away"] = self.away()
        df["season"] = self.season()
        df["week"] = self.week()
        feats = sportsref.nfl.pbp.expand_details(df)

        # add team and opp columns by iterating through rows
        df = sportsref.nfl.pbp._add_team_columns(feats)
        # add WPA column (requires diff, can't be done row-wise)
        df["home_wpa"] = df.home_wp.diff()
        # lag score columns, fill in 0-0 to start
        for col in ("home_wp", "pbp_score_hm", "pbp_score_aw"):
            if col in df.columns:
                df[col] = df[col].shift(1)
        df.loc[0, ["pbp_score_hm", "pbp_score_aw"]] = 0
        # fill in WP NaN's
        df.home_wp.fillna(method="ffill", inplace=True)
        # fix first play border after diffing/shifting for WP and WPA
        firstPlaysOfGame = df[df.secsElapsed == 0].index
        line = self.line()
        for i in firstPlaysOfGame:
            initwp = sportsref.nfl.winProb.initialWinProb(line)
            df.loc[i, "home_wp"] = initwp
            df.loc[i, "home_wpa"] = df.loc[i + 1, "home_wp"] - initwp
        # fix last play border after diffing/shifting for WP and WPA
        lastPlayIdx = df.index[-1]
        lastPlayWP = df.loc[lastPlayIdx, "home_wp"]
        # if a tie, final WP is 50%; otherwise, determined by winner
        winner = self.winner()
        finalWP = 50.0 if pd.isnull(winner) else (winner == self.home()) * 100.0
        df.loc[lastPlayIdx, "home_wpa"] = finalWP - lastPlayWP
        # fix WPA for timeouts and plays after timeouts
        timeouts = df[df.isTimeout].index
        for to in timeouts:
            df.loc[to, "home_wpa"] = 0.0
            if to + 2 in df.index:
                wpa = df.loc[to + 2, "home_wp"] - df.loc[to + 1, "home_wp"]
            else:
                wpa = finalWP - df.loc[to + 1, "home_wp"]
            df.loc[to + 1, "home_wpa"] = wpa
        # add team-related features to DataFrame
        df = sportsref.nfl.pbp._add_team_features(df)
        # fill distToGoal NaN's
        df["distToGoal"] = np.where(df.isKickoff, 65, df.distToGoal)
        df.distToGoal.fillna(method="bfill", inplace=True)
        df.distToGoal.fillna(method="ffill", inplace=True)  # for last play

        return df

    @sportsref.decorators.memoize
    def ref_info(self):
        """Gets a dictionary of ref positions and the ref IDs of the refs for
        that game.

        :returns: A dictionary of ref positions and IDs.
        """
        doc = self.get_doc()
        table = doc("table#officials")
        return sportsref.utils.parse_info_table(table)

    @sportsref.decorators.memoize
    def player_stats(self):
        """Gets the stats for offense, defense, returning, and kicking of
        individual players in the game.
        :returns: A DataFrame containing individual player stats.
        """
        doc = self.get_doc()
        tableIDs = ("player_offense", "player_defense", "returns", "kicking")
        dfs = []
        for tID in tableIDs:
            table = doc("table#{}".format(tID))
            dfs.append(sportsref.utils.parse_table(table))
        dfs = [df for df in dfs if not df.empty]
        df = functools.reduce(
            lambda x, y: pd.merge(
                x, y, how="outer", on=list(set(x.columns) & set(y.columns))
            ),
            dfs,
        ).reset_index(drop=True)
        return df

    @sportsref.decorators.memoize
    def snap_counts(self):
        """Gets the snap counts for both teams' players and returns them in a
        DataFrame. Note: only goes back to 2012.

        :returns: DataFrame of snap count data
        """
        # TODO: combine duplicate players, see 201312150mia - ThomDa03
        doc = self.get_doc()
        table_ids = ("vis_snap_counts", "home_snap_counts")
        tms = (self.away(), self.home())
        df = pd.concat(
            [
                sportsref.utils.parse_table(doc("table#{}".format(table_id))).assign(
                    is_home=bool(i), team=tms[i], opp=tms[i * -1 + 1]
                )
                for i, table_id in enumerate(table_ids)
            ]
        )
        if df.empty:
            return df
        return df.set_index("player_id")



import datetime
import re

import numpy as np
import pandas as pd
from pyquery import PyQuery as pq

import sportsref

CLOCK_REGEX = re.compile(r"(\d+):(\d+)\.(\d+)")


class BoxScore(object, metaclass=sportsref.decorators.Cached):
    def __init__(self, boxscore_id):
        self.boxscore_id = boxscore_id

    def __eq__(self, other):
        return self.boxscore_id == other.boxscore_id

    def __hash__(self):
        return hash(self.boxscore_id)

    def __repr__(self):
        return f"BoxScore({self.boxscore_id})"

    @sportsref.decorators.memoize
    def get_main_doc(self):
        url = f"{sportsref.nba.BASE_URL}/boxscores/{self.boxscore_id}.html"
        doc = pq(sportsref.utils.get_html(url))
        return doc

    @sportsref.decorators.memoize
    def get_subpage_doc(self, page):
        url = f"{sportsref.nba.BASE_URL}/boxscores/{page}/{self.boxscore_id}.html"
        doc = pq(sportsref.utils.get_html(url))
        return doc

    @sportsref.decorators.memoize
    def date(self):
        """Returns the date of the game. See Python datetime.date documentation
        for more.
        :returns: A datetime.date object with year, month, and day attributes.
        """
        match = re.match(r"(\d{4})(\d{2})(\d{2})", self.boxscore_id)
        year, month, day = list(map(int, match.groups()))
        return datetime.date(year=year, month=month, day=day)

    @sportsref.decorators.memoize
    def weekday(self):
        days = [
            "Monday",
            "Tuesday",
            "Wednesday",
            "Thursday",
            "Friday",
            "Saturday",
            "Sunday",
        ]
        date = self.date()
        wd = date.weekday()
        return days[wd]

    @sportsref.decorators.memoize
    def linescore(self):
        """Returns the linescore for the game as a DataFrame."""
        doc = self.get_main_doc()
        table = doc("table#line_score")
        df = sportsref.utils.parse_table(table)
        df.index = ["away", "home"]
        return df

    @sportsref.decorators.memoize
    def home(self):
        """Returns home team ID.
        :returns: 3-character string representing home team's ID.
        """
        linescore = self.linescore()
        return linescore.loc["home", "team_id"]

    @sportsref.decorators.memoize
    def away(self):
        """Returns away team ID.
        :returns: 3-character string representing away team's ID.
        """
        linescore = self.linescore()
        return linescore.loc["away", "team_id"]

    @sportsref.decorators.memoize
    def home_score(self):
        """Returns score of the home team.
        :returns: int of the home score.
        """
        linescore = self.linescore()
        return linescore.loc["home", "T"]

    @sportsref.decorators.memoize
    def away_score(self):
        """Returns score of the away team.
        :returns: int of the away score.
        """
        linescore = self.linescore()
        return linescore.loc["away", "T"]

    @sportsref.decorators.memoize
    def winner(self):
        """Returns the team ID of the winning team. Returns NaN if a tie."""
        hmScore = self.home_score()
        awScore = self.away_score()
        if hmScore > awScore:
            return self.home()
        elif hmScore < awScore:
            return self.away()
        else:
            return None

    @sportsref.decorators.memoize
    def season(self):
        """
        Returns the year ID of the season in which this game took place.

        :returns: An int representing the year of the season.
        """
        d = self.date()
        if d.month >= 9:
            return d.year + 1
        else:
            return d.year

    def _get_player_stats(self, table_id_fmt):
        """Returns a DataFrame of player stats from the game (either basic or
        advanced, depending on the argument.

        :param table_id_fmt: Format string for str.format with a placeholder
            for the team ID (e.g. 'box-{}-game-basic')
        :returns: DataFrame of player stats
        """

        # get data
        doc = self.get_main_doc()
        tms = self.away(), self.home()
        team_table_ids = [table_id_fmt.format(tm.upper()) for tm in tms]
        tables = [doc(f"table#{table_id}") for table_id in team_table_ids]
        dfs = [sportsref.utils.parse_table(table) for table in tables]

        # clean data and add features
        for i, (tm, df) in enumerate(zip(tms, dfs)):
            no_time = df["mp"] == 0
            stat_cols = [
                col for col, dtype in list(df.dtypes.items()) if dtype != "object"
            ]
            df.loc[no_time, stat_cols] = 0
            df["team_id"] = tm
            df["is_home"] = i == 1
            df["is_starter"] = [p < 5 for p in range(df.shape[0])]
            df.drop_duplicates(subset="player_id", keep="first", inplace=True)

        return pd.concat(dfs)

    @sportsref.decorators.memoize
    def basic_stats(self):
        """Returns a DataFrame of basic player stats from the game."""
        return self._get_player_stats("box-{}-game-basic")

    @sportsref.decorators.memoize
    def advanced_stats(self):
        """Returns a DataFrame of advanced player stats from the game."""
        return self._get_player_stats("box-{}-game-advanced")

    @sportsref.decorators.memoize
    def pbp(self, dense_lineups=False, sparse_lineups=False):
        """Returns a dataframe of the play-by-play data from the game.

        :param dense_lineups: If True, adds 10 columns containing the names of
            the players on the court. Defaults to False.
        :param sparse_lineups: If True, adds binary columns denoting whether a
            given player is in the game at the time of a pass. Defaults to
            False.
        :returns: pandas DataFrame of play-by-play. Similar to GPF.
        """
        try:
            doc = self.get_subpage_doc("pbp")
        except Exception:
            raise ValueError(
                f"Error fetching PBP subpage for boxscore {self.boxscore_id}"
            )

        table = doc("table#pbp")
        trs = [
            tr
            for tr in list(table("tr").items())
            if (
                not tr.attr["class"]
                or (tr.attr["id"] and tr.attr["id"].startswith("q"))
            )
        ]
        rows = [tr.children("td") for tr in trs]
        n_rows = len(trs)
        data = []
        cur_qtr = 0

        for i in range(n_rows):
            tr = trs[i]
            row = rows[i]
            play = {}

            # increment cur_qtr when we hit a new quarter
            if tr.attr["id"] and tr.attr["id"].startswith("q"):
                assert int(tr.attr["id"][1:]) == cur_qtr + 1
                cur_qtr += 1
                continue

            # add time of play to entry
            clock_str = row.eq(0).text()
            mins, secs, tenths = list(
                map(int, re.match(CLOCK_REGEX, clock_str).groups())
            )
            secs_in_period = 12 * 60 * min(cur_qtr, 4) + 5 * 60 * (
                cur_qtr - 4 if cur_qtr > 4 else 0
            )
            secs_elapsed = secs_in_period - (60 * mins + secs + 0.1 * tenths)
            play["secs_elapsed"] = secs_elapsed
            play["clock_str"] = clock_str
            play["quarter"] = cur_qtr

            # handle single play description
            # ex: beginning/end of quarter, jump ball
            if row.length == 2:
                desc = row.eq(1)
                # handle jump balls
                if desc.text().lower().startswith("jump ball: "):
                    play["is_jump_ball"] = True
                    jump_ball_str = sportsref.utils.flatten_links(desc)
                    play.update(
                        sportsref.nba.pbp.parse_play(
                            self.boxscore_id, jump_ball_str, is_home=None
                        )
                    )
                # ignore rows marking beginning/end of quarters
                elif desc.text().lower().startswith(
                    "start of "
                ) or desc.text().lower().startswith("end of "):
                    continue
                # if another case, log and continue
                else:
                    if not desc.text().lower().startswith("end of "):
                        print(
                            f"{self.boxscore_id}, Q{cur_qtr}, {clock_str} other case: {desc.text()}"
                        )
                    continue

            # handle team play description
            # ex: shot, turnover, rebound, foul, sub, etc.
            elif row.length == 6:
                aw_desc, hm_desc = row.eq(1), row.eq(5)
                is_hm_play = bool(hm_desc.text())
                desc = hm_desc if is_hm_play else aw_desc
                desc = sportsref.utils.flatten_links(desc)
                # parse the play
                new_play = sportsref.nba.pbp.parse_play(
                    self.boxscore_id, desc, is_hm_play
                )
                if not new_play:
                    continue
                elif isinstance(new_play, list):
                    # this happens when a row needs to be expanded to 2 rows;
                    # ex: double personal foul -> two PF rows

                    # first, update and append the first row
                    orig_play = dict(play)
                    play.update(new_play[0])
                    data.append(play)
                    # second, set up the second row to be appended below
                    play = orig_play
                    new_play = new_play[1]
                elif new_play.get("is_error"):
                    print(f"can't parse: {desc}, boxscore: {self.boxscore_id}")
                    # import pdb; pdb.set_trace()
                play.update(new_play)

            # otherwise, I don't know what this was
            else:
                raise Exception(f"don't know how to handle row of length {row.length}")

            data.append(play)

        # convert to DataFrame and clean columns
        df = pd.DataFrame.from_records(data)
        df.sort_values("secs_elapsed", inplace=True, kind="mergesort")
        df = sportsref.nba.pbp.clean_features(df)

        # add columns for home team, away team, boxscore_id, date
        away, home = self.away(), self.home()
        df["home"] = home
        df["away"] = away
        df["boxscore_id"] = self.boxscore_id
        df["season"] = self.season()
        date = self.date()
        df["year"] = date.year
        df["month"] = date.month
        df["day"] = date.day

        def _clean_rebs(df):
            df.reset_index(drop=True, inplace=True)
            no_reb_after = (
                ((df.fta_num < df.tot_fta) | df.is_ftm | df.get("is_tech_fta", False))
                .shift(1)
                .fillna(False)
            )
            no_reb_before = ((df.fta_num == df.tot_fta)).shift(-1).fillna(False)
            se_end_qtr = df.loc[df.clock_str == "0:00.0", "secs_elapsed"].unique()
            no_reb_when = df.secs_elapsed.isin(se_end_qtr)
            drop_mask = (df.rebounder == "Team") & (
                no_reb_after | no_reb_before | no_reb_when
            )
            df.drop(df.loc[drop_mask].index, axis=0, inplace=True)
            df.reset_index(drop=True, inplace=True)
            return df

        # get rid of 'rebounds' after FTM, non-final FTA, or tech FTA
        df = _clean_rebs(df)

        # track possession number for each possession
        # TODO: see 201604130PHO, secs_elapsed == 2756
        # things that end a poss:
        # FGM, dreb, TO, end of Q, made last FT, lost jump ball,
        # def goaltending, shot clock violation
        new_poss = (df.off_team == df.home).diff().fillna(False)
        # def rebound considered part of the new possession
        df["poss_id"] = np.cumsum(new_poss) + df.is_dreb
        # create poss_id with rebs -> new possessions for granular groupbys
        poss_id_reb = np.cumsum(new_poss | df.is_reb)

        # make sure plays with the same clock time are in the right order
        # TODO: make sort_cols depend on what cols are in the play?
        # or combine related plays, like and-1 shot and foul
        # issues come up with FGA after timeout in 201604130LAL
        # issues come up with PF between FGA and DREB in 201604120SAS
        sort_cols = [
            col
            for col in [
                "is_reb",
                "is_fga",
                "is_pf",
                "is_tech_foul",
                "is_ejection",
                "is_tech_fta",
                "is_timeout",
                "is_pf_fta",
                "fta_num",
                "is_viol",
                "is_to",
                "is_jump_ball",
                "is_sub",
            ]
            if col in df.columns
        ]
        asc_true = ["fta_num"]
        ascend = [(col in asc_true) for col in sort_cols]
        for label, group in df.groupby([df.secs_elapsed, poss_id_reb]):
            if len(group) > 1:
                df.loc[group.index, :] = group.sort_values(
                    sort_cols, ascending=ascend, kind="mergesort"
                ).values

        # 2nd pass: get rid of 'rebounds' after FTM, non-final FTA, etc.
        df = _clean_rebs(df)

        # makes sure off/def and poss_id are correct for subs after rearranging
        # some possessions above
        df.loc[df["is_sub"], ["off_team", "def_team", "poss_id"]] = np.nan
        df.off_team.fillna(method="bfill", inplace=True)
        df.def_team.fillna(method="bfill", inplace=True)
        df.poss_id.fillna(method="bfill", inplace=True)
        # make off_team and def_team NaN for jump balls
        if "is_jump_ball" in df.columns:
            df.loc[df["is_jump_ball"], ["off_team", "def_team"]] = np.nan

        # make sure 'off_team' is always the team shooting FTs, even on techs
        # (impt for keeping track of the score)
        if "is_tech_fta" in df.columns:
            tech_fta = df["is_tech_fta"]
            df.loc[tech_fta, "off_team"] = df.loc[tech_fta, "fta_team"]
            df.loc[tech_fta, "def_team"] = np.where(
                df.loc[tech_fta, "off_team"] == home, away, home
            )
        df.drop("fta_team", axis=1, inplace=True)
        # redefine poss_id_reb
        new_poss = (df.off_team == df.home).diff().fillna(False)
        poss_id_reb = np.cumsum(new_poss | df.is_reb)

        # get rid of redundant subs
        for (se, tm, pnum), group in df[df.is_sub].groupby(
            [df.secs_elapsed, df.sub_team, poss_id_reb]
        ):
            if len(group) > 1:
                sub_in = set()
                sub_out = set()
                # first, figure out who's in and who's out after subs
                for i, row in group.iterrows():
                    if row["sub_in"] in sub_out:
                        sub_out.remove(row["sub_in"])
                    else:
                        sub_in.add(row["sub_in"])
                    if row["sub_out"] in sub_in:
                        sub_in.remove(row["sub_out"])
                    else:
                        sub_out.add(row["sub_out"])
                assert len(sub_in) == len(sub_out)
                # second, add those subs
                n_subs = len(sub_in)
                for idx, p_in, p_out in zip(group.index[:n_subs], sub_in, sub_out):
                    assert df.loc[idx, "is_sub"]
                    df.loc[idx, "sub_in"] = p_in
                    df.loc[idx, "sub_out"] = p_out
                    df.loc[idx, "sub_team"] = tm
                    df.loc[idx, "detail"] = f"{p_in} enters the game for {p_out}"
                # third, if applicable, remove old sub entries when there are
                # redundant subs
                n_extra = len(group) - len(sub_in)
                if n_extra:
                    extra_idxs = group.index[-n_extra:]
                    df.drop(extra_idxs, axis=0, inplace=True)

        df.reset_index(drop=True, inplace=True)

        # add column for pts and score
        df["pts"] = df["is_ftm"] + 2 * df["is_fgm"] + (df["is_fgm"] & df["is_three"])
        df["hm_pts"] = np.where(df.off_team == df.home, df.pts, 0)
        df["aw_pts"] = np.where(df.off_team == df.away, df.pts, 0)
        df["hm_score"] = np.cumsum(df["hm_pts"])
        df["aw_score"] = np.cumsum(df["aw_pts"])

        # more helpful columns
        # "play" is differentiated from "poss" by counting OReb as new play
        # "plays" end with non-and1 FGA, TO, last non-tech FTA, or end of qtr
        # (or double lane viol)
        new_qtr = df.quarter.diff().shift(-1).fillna(False).astype(bool)  # noqa
        and1 = (  # noqa
            df.is_fgm
            & df.is_pf.shift(-1).fillna(False)
            & df.is_fta.shift(-2).fillna(False)
            & ~df.secs_elapsed.diff().shift(-1).fillna(False).astype(bool)
        )
        double_lane = df.get("viol_type") == "double lane"  # noqa

        new_play = df.eval(
            "(is_fga & ~(@and1)) | is_to | @new_qtr |"
            "(is_fta & ~is_tech_fta & fta_num == tot_fta) |"
            "@double_lane"
        )
        df["play_id"] = np.cumsum(new_play).shift(1).fillna(0)
        df["hm_off"] = df.off_team == df.home

        # get lineup data
        if dense_lineups:
            df = pd.concat((df, sportsref.nba.pbp.get_dense_lineups(df)), axis=1)
        if sparse_lineups:
            df = pd.concat((df, sportsref.nba.pbp.get_sparse_lineups(df)), axis=1)

        # TODO: add shot clock as a feature

        return df



from . import boxscores
from . import pbp
from . import seasons
from . import teams

from .boxscores import BoxScore
from .seasons import Season
from .teams import Team
from .players import Player

BASE_URL = "http://www.basketball-reference.com"

__all__ = [
    "BASE_URL",
    "boxscores",
    "BoxScore",
    "pbp",
    "seasons",
    "Season",
    "teams",
    "Team",
    "players",
    "Player",
]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sportsref', 'sportsref.nba', 'sportsref.nfl', 'sportsref.nfl.finders']

package_data = \
{'': ['*']}

install_requires = \
['mementos>=1.3.1,<2.0.0',
 'numexpr>=2.7.1,<3.0.0',
 'numpy>=1.19.4,<2.0.0',
 'pandas>=1.1.4,<2.0.0',
 'pyquery>=1.4.3,<2.0.0',
 'requests>=2.25.0,<3.0.0']

setup_kwargs = {
    'name': 'sportsref',
    'version': '0.12.0',
    'description': '',
    'long_description': None,
    'author': 'Matt Goldberg',
    'author_email': 'matt.goldberg7@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)



import datetime
import re
import urllib.parse

from pyquery import PyQuery as pq

import sportsref

__all__ = ["Player"]


class Player(object, metaclass=sportsref.decorators.Cached):
    def __init__(self, player_id):
        self.player_id = player_id
        self.mainURL = (sportsref.nfl.BASE_URL + "/players/{0[0]}/{0}.htm").format(
            self.player_id
        )

    def __eq__(self, other):
        return self.player_id == other.player_id

    def __hash__(self):
        return hash(self.player_id)

    def __repr__(self):
        return "Player({})".format(self.player_id)

    def __str__(self):
        return self.name()

    def __reduce__(self):
        return Player, (self.player_id,)

    def _subpage_url(self, page, year=None):
        # if no year, return career version
        if year is None:
            return urllib.parse.urljoin(
                self.mainURL, "{}/{}/".format(self.player_id, page)
            )
        # otherwise, return URL for a given year
        else:
            return urllib.parse.urljoin(
                self.mainURL, "{}/{}/{}/".format(self.player_id, page, year)
            )

    @sportsref.decorators.memoize
    def get_doc(self):
        doc = pq(sportsref.utils.get_html(self.mainURL))
        return doc

    @sportsref.decorators.memoize
    def name(self):
        doc = self.get_doc()
        name = doc("div#meta h1:first").text()
        return name

    @sportsref.decorators.memoize
    def age(self, year, month=9, day=1):
        doc = self.get_doc()
        span = doc("div#meta span#necro-birth")
        birthstring = span.attr("data-birth")
        try:
            dateargs = re.match(r"(\d{4})\-(\d{2})\-(\d{2})", birthstring).groups()
            dateargs = list(map(int, dateargs))
            birthDate = datetime.date(*dateargs)
            delta = datetime.date(year=year, month=month, day=day) - birthDate
            age = delta.days / 365
            return age
        except Exception:
            return None

    @sportsref.decorators.memoize
    def position(self):
        doc = self.get_doc()
        rawText = (
            doc("div#meta p").filter(lambda i, e: "Position" in e.text_content()).text()
        )
        rawPos = re.search(r"Position\W*(\S+)", rawText, re.I).group(1)
        allPositions = rawPos.split("-")
        # right now, returning just the primary position for those with
        # multiple positions
        return allPositions[0]

    @sportsref.decorators.memoize
    def height(self):
        doc = self.get_doc()
        rawText = doc('div#meta p span[itemprop="height"]').text()
        try:
            feet, inches = list(map(int, rawText.split("-")))
            return feet * 12 + inches
        except ValueError:
            return None

    @sportsref.decorators.memoize
    def weight(self):
        doc = self.get_doc()
        rawText = doc('div#meta p span[itemprop="weight"]').text()
        try:
            weight = re.match(r"(\d+)lb", rawText, re.I).group(1)
            return int(weight)
        except AttributeError:
            return None

    @sportsref.decorators.memoize
    def hand(self):
        doc = self.get_doc()
        try:
            rawText = (
                doc("div#meta p")
                .filter(lambda i, e: "Throws" in e.text_content())
                .text()
            )
            rawHand = re.search(r"Throws\W+(\S+)", rawText, re.I).group(1)
        except AttributeError:
            return None
        return rawHand[0]  # 'L' or 'R'

    @sportsref.decorators.memoize
    def current_team(self):
        doc = self.get_doc()
        team = doc("div#meta p").filter(lambda i, e: "Team" in e.text_content())
        text = sportsref.utils.flatten_links(team)
        try:
            m = re.match(r"Team: (\w{3})", text)
            return m.group(1)
        except Exception:
            return None

    @sportsref.decorators.memoize
    def draft_pick(self):
        doc = self.get_doc()
        rawDraft = (
            doc("div#meta p").filter(lambda i, e: "Draft" in e.text_content()).text()
        )
        m = re.search(r"Draft.*? round \((\d+).*?overall\)", rawDraft, re.I)
        # if not drafted or taken in supplemental draft, return NaN
        if m is None or "Supplemental" in rawDraft:
            return None
        else:
            return int(m.group(1))

    @sportsref.decorators.memoize
    def draft_class(self):
        doc = self.get_doc()
        rawDraft = (
            doc("div#meta p").filter(lambda i, e: "Draft" in e.text_content()).text()
        )
        m = re.search(r"Draft.*?of the (\d{4}) NFL", rawDraft, re.I)
        if not m:
            return None
        else:
            return int(m.group(1))

    @sportsref.decorators.memoize
    def draft_team(self):
        doc = self.get_doc()
        rawDraft = doc("div#meta p").filter(lambda i, e: "Draft" in e.text_content())
        try:
            draftStr = sportsref.utils.flatten_links(rawDraft)
            m = re.search(r"Draft\W+(\w+)", draftStr)
            return m.group(1)
        except Exception:
            return None

    @sportsref.decorators.memoize
    def college(self):
        doc = self.get_doc()
        rawText = doc("div#meta p").filter(lambda i, e: "College" in e.text_content())
        cleanedText = sportsref.utils.flatten_links(rawText)
        college = re.search(r"College:\s*(\S+)", cleanedText).group(1)
        return college

    @sportsref.decorators.memoize
    def high_school(self):
        doc = self.get_doc()
        rawText = doc("div#meta p").filter(
            lambda i, e: "High School" in e.text_content()
        )
        cleanedText = sportsref.utils.flatten_links(rawText)
        hs = re.search(r"High School:\s*(\S+)", cleanedText).group(1)
        return hs

    @sportsref.decorators.memoize
    @sportsref.decorators.kind_rpb(include_type=True)
    def gamelog(self, year=None, kind="R"):
        """Gets the career gamelog of the given player.
        :kind: One of 'R', 'P', or 'B' (for regular season, playoffs, or both).
        Case-insensitive; defaults to 'R'.
        :year: The year for which the gamelog should be returned; if None,
        return entire career gamelog. Defaults to None.
        :returns: A DataFrame with the player's career gamelog.
        """
        url = self._subpage_url("gamelog", None)  # year is filtered later
        doc = pq(sportsref.utils.get_html(url))
        table = doc("table#stats") if kind == "R" else doc("table#stats_playoffs")
        df = sportsref.utils.parse_table(table)
        if year is not None:
            df = df.query("year == @year").reset_index(drop=True)
        return df

    @sportsref.decorators.memoize
    @sportsref.decorators.kind_rpb(include_type=True)
    def passing(self, kind="R"):
        """Gets yearly passing stats for the player.

        :kind: One of 'R', 'P', or 'B'. Case-insensitive; defaults to 'R'.
        :returns: Pandas DataFrame with passing stats.
        """
        doc = self.get_doc()
        table = doc("table#passing") if kind == "R" else doc("table#passing_playoffs")
        df = sportsref.utils.parse_table(table)
        return df

    @sportsref.decorators.memoize
    @sportsref.decorators.kind_rpb(include_type=True)
    def rushing_and_receiving(self, kind="R"):
        """Gets yearly rushing/receiving stats for the player.

        :kind: One of 'R', 'P', or 'B'. Case-insensitive; defaults to 'R'.
        :returns: Pandas DataFrame with rushing/receiving stats.
        """
        doc = self.get_doc()
        table = (
            doc("table#rushing_and_receiving")
            if kind == "R"
            else doc("table#rushing_and_receiving_playoffs")
        )
        if not table:
            table = (
                doc("table#receiving_and_rushing")
                if kind == "R"
                else doc("table#receiving_and_rushing_playoffs")
            )
        df = sportsref.utils.parse_table(table)
        return df

    @sportsref.decorators.memoize
    @sportsref.decorators.kind_rpb(include_type=True)
    def defense(self, kind="R"):
        """Gets yearly defense stats for the player (also has AV stats for OL).

        :kind: One of 'R', 'P', or 'B'. Case-insensitive; defaults to 'R'.
        :returns: Pandas DataFrame with rushing/receiving stats.
        """
        doc = self.get_doc()
        table = doc("table#defense") if kind == "R" else doc("table#defense_playoffs")
        df = sportsref.utils.parse_table(table)
        return df

    def _plays(self, year, play_type, expand_details):
        """Returns a DataFrame of plays for a given year for a given play type
        (like rushing, receiving, or passing).

        :year: The year for the season.
        :play_type: A type of play for which there are plays (as of this
        writing, either "passing", "rushing", or "receiving")
        :expand_details: Bool for whether PBP should be parsed.
        :returns: A DataFrame of plays, each row is a play. Returns None if
        there were no such plays in that year.
        """
        url = self._subpage_url("{}-plays".format(play_type), year)
        doc = pq(sportsref.utils.get_html(url))
        table = doc("table#all_plays")
        if table:
            if expand_details:
                plays = sportsref.nfl.pbp.expand_details(
                    sportsref.utils.parse_table(table), detailCol="description"
                )
                return plays
            else:
                return sportsref.utils.parse_table(table)
        else:
            return None

    @sportsref.decorators.memoize
    def passing_plays(self, year, expand_details=True):
        """Returns a pbp DataFrame of a player's passing plays in a season.

        :year: The year for the season.
        :expand_details: bool for whether PBP should be parsed.
        :returns: A DataFrame of stats, each row is a play.
        """
        return self._plays(year, "passing", expand_details)

    @sportsref.decorators.memoize
    def rushing_plays(self, year, expand_details=True):
        """Returns a pbp DataFrame of a player's rushing plays in a season.

        :year: The year for the season.
        :expand_details: bool for whether PBP should be parsed.
        :returns: A DataFrame of stats, each row is a play.
        """
        return self._plays(year, "rushing", expand_details)

    @sportsref.decorators.memoize
    def receiving_plays(self, year, expand_details=True):
        """Returns a pbp DataFrame of a player's receiving plays in a season.

        :year: The year for the season.
        :expand_details: bool for whether PBP should be parsed.
        :returns: A DataFrame of stats, each row is a play.
        """
        return self._plays(year, "receiving", expand_details)

    @sportsref.decorators.memoize
    def splits(self, year=None):
        """Returns a DataFrame of splits data for a player-year.

        :year: The year for the season in question. If None, returns career
        splits.
        :returns: A DataFrame of splits data.
        """
        # get the table
        url = self._subpage_url("splits", year)
        doc = pq(sportsref.utils.get_html(url))
        table = doc("table#stats")
        df = sportsref.utils.parse_table(table)
        # cleaning the data
        if not df.empty:
            df.split_id.fillna(method="ffill", inplace=True)
        return df

    @sportsref.decorators.memoize
    def advanced_splits(self, year=None):
        """Returns a DataFrame of advanced splits data for a player-year. Note:
            only go back to 2012.

        :year: The year for the season in question. If None, returns career
        advanced splits.
        :returns: A DataFrame of advanced splits data.
        """
        # get the table
        url = self._subpage_url("splits", year)
        doc = pq(sportsref.utils.get_html(url))
        table = doc("table#advanced_splits")
        df = sportsref.utils.parse_table(table)
        # cleaning the data
        if not df.empty:
            df.split_type.fillna(method="ffill", inplace=True)
        return df

    @sportsref.decorators.memoize
    def _simple_year_award(self, award_id):
        """Template for simple award functions that simply list years, such as
        pro bowls and first-team all pro.

        :award_id: The div ID that is appended to "leaderboard_" in selecting
        the table's div.
        :returns: List of years for the award.
        """
        doc = self.get_doc()
        table = doc("div#leaderboard_{} table".format(award_id))
        return list(map(int, sportsref.utils.parse_awards_table(table)))

    def pro_bowls(self):
        """Returns a list of years in which the player made the Pro Bowl."""
        return self._simple_year_award("pro_bowls")

    def first_team_all_pros(self):
        """Returns a list of years in which the player made 1st-Tm All Pro."""
        return self._simple_year_award("all_pro")

    # TODO: other awards like MVP, OPOY, DPOY, NFL Top 100, etc.
